from cognite.powerops.client._generated.v1._api.alert import AlertAPI
from cognite.powerops.client._generated.v1._api.alert_query import AlertQueryAPI
from cognite.powerops.client._generated.v1._api.benchmarking_calculation_input import BenchmarkingCalculationInputAPI
from cognite.powerops.client._generated.v1._api.benchmarking_calculation_input_query import BenchmarkingCalculationInputQueryAPI
from cognite.powerops.client._generated.v1._api.benchmarking_calculation_input_shop_results import BenchmarkingCalculationInputShopResultsAPI
from cognite.powerops.client._generated.v1._api.benchmarking_calculation_output import BenchmarkingCalculationOutputAPI
from cognite.powerops.client._generated.v1._api.benchmarking_calculation_output_alerts import BenchmarkingCalculationOutputAlertsAPI
from cognite.powerops.client._generated.v1._api.benchmarking_calculation_output_benchmarking_results import BenchmarkingCalculationOutputBenchmarkingResultsAPI
from cognite.powerops.client._generated.v1._api.benchmarking_calculation_output_query import BenchmarkingCalculationOutputQueryAPI
from cognite.powerops.client._generated.v1._api.benchmarking_configuration_day_ahead import BenchmarkingConfigurationDayAheadAPI
from cognite.powerops.client._generated.v1._api.benchmarking_configuration_day_ahead_assets_per_shop_model import BenchmarkingConfigurationDayAheadAssetsPerShopModelAPI
from cognite.powerops.client._generated.v1._api.benchmarking_configuration_day_ahead_bid_configurations import BenchmarkingConfigurationDayAheadBidConfigurationsAPI
from cognite.powerops.client._generated.v1._api.benchmarking_configuration_day_ahead_query import BenchmarkingConfigurationDayAheadQueryAPI
from cognite.powerops.client._generated.v1._api.benchmarking_production_obligation_day_ahead import BenchmarkingProductionObligationDayAheadAPI
from cognite.powerops.client._generated.v1._api.benchmarking_production_obligation_day_ahead_query import BenchmarkingProductionObligationDayAheadQueryAPI
from cognite.powerops.client._generated.v1._api.benchmarking_production_obligation_day_ahead_time_series import BenchmarkingProductionObligationDayAheadTimeSeriesAPI
from cognite.powerops.client._generated.v1._api.benchmarking_result_day_ahead import BenchmarkingResultDayAheadAPI
from cognite.powerops.client._generated.v1._api.benchmarking_result_day_ahead_alerts import BenchmarkingResultDayAheadAlertsAPI
from cognite.powerops.client._generated.v1._api.benchmarking_result_day_ahead_query import BenchmarkingResultDayAheadQueryAPI
from cognite.powerops.client._generated.v1._api.benchmarking_shop_case import BenchmarkingShopCaseAPI
from cognite.powerops.client._generated.v1._api.benchmarking_shop_case_query import BenchmarkingShopCaseQueryAPI
from cognite.powerops.client._generated.v1._api.benchmarking_shop_case_shop_files import BenchmarkingShopCaseShopFilesAPI
from cognite.powerops.client._generated.v1._api.benchmarking_task_dispatcher_input_day_ahead import BenchmarkingTaskDispatcherInputDayAheadAPI
from cognite.powerops.client._generated.v1._api.benchmarking_task_dispatcher_input_day_ahead_query import BenchmarkingTaskDispatcherInputDayAheadQueryAPI
from cognite.powerops.client._generated.v1._api.benchmarking_task_dispatcher_output_day_ahead import BenchmarkingTaskDispatcherOutputDayAheadAPI
from cognite.powerops.client._generated.v1._api.benchmarking_task_dispatcher_output_day_ahead_alerts import BenchmarkingTaskDispatcherOutputDayAheadAlertsAPI
from cognite.powerops.client._generated.v1._api.benchmarking_task_dispatcher_output_day_ahead_benchmarking_sub_tasks import BenchmarkingTaskDispatcherOutputDayAheadBenchmarkingSubTasksAPI
from cognite.powerops.client._generated.v1._api.benchmarking_task_dispatcher_output_day_ahead_query import BenchmarkingTaskDispatcherOutputDayAheadQueryAPI
from cognite.powerops.client._generated.v1._api.bid_configuration_day_ahead import BidConfigurationDayAheadAPI
from cognite.powerops.client._generated.v1._api.bid_configuration_day_ahead_partials import BidConfigurationDayAheadPartialsAPI
from cognite.powerops.client._generated.v1._api.bid_configuration_day_ahead_query import BidConfigurationDayAheadQueryAPI
from cognite.powerops.client._generated.v1._api.bid_document_afrr import BidDocumentAFRRAPI
from cognite.powerops.client._generated.v1._api.bid_document_afrr_alerts import BidDocumentAFRRAlertsAPI
from cognite.powerops.client._generated.v1._api.bid_document_afrr_bids import BidDocumentAFRRBidsAPI
from cognite.powerops.client._generated.v1._api.bid_document_afrr_query import BidDocumentAFRRQueryAPI
from cognite.powerops.client._generated.v1._api.bid_document import BidDocumentAPI
from cognite.powerops.client._generated.v1._api.bid_document_alerts import BidDocumentAlertsAPI
from cognite.powerops.client._generated.v1._api.bid_document_day_ahead import BidDocumentDayAheadAPI
from cognite.powerops.client._generated.v1._api.bid_document_day_ahead_alerts import BidDocumentDayAheadAlertsAPI
from cognite.powerops.client._generated.v1._api.bid_document_day_ahead_partials import BidDocumentDayAheadPartialsAPI
from cognite.powerops.client._generated.v1._api.bid_document_day_ahead_query import BidDocumentDayAheadQueryAPI
from cognite.powerops.client._generated.v1._api.bid_document_query import BidDocumentQueryAPI
from cognite.powerops.client._generated.v1._api.bid_matrix import BidMatrixAPI
from cognite.powerops.client._generated.v1._api.bid_matrix_information import BidMatrixInformationAPI
from cognite.powerops.client._generated.v1._api.bid_matrix_information_alerts import BidMatrixInformationAlertsAPI
from cognite.powerops.client._generated.v1._api.bid_matrix_information_linked_time_series import BidMatrixInformationLinkedTimeSeriesAPI
from cognite.powerops.client._generated.v1._api.bid_matrix_information_query import BidMatrixInformationQueryAPI
from cognite.powerops.client._generated.v1._api.bid_matrix_information_underlying_bid_matrices import BidMatrixInformationUnderlyingBidMatricesAPI
from cognite.powerops.client._generated.v1._api.bid_matrix_query import BidMatrixQueryAPI
from cognite.powerops.client._generated.v1._api.bid_row import BidRowAPI
from cognite.powerops.client._generated.v1._api.bid_row_alerts import BidRowAlertsAPI
from cognite.powerops.client._generated.v1._api.bid_row_query import BidRowQueryAPI
from cognite.powerops.client._generated.v1._api.date_specification import DateSpecificationAPI
from cognite.powerops.client._generated.v1._api.date_specification_query import DateSpecificationQueryAPI
from cognite.powerops.client._generated.v1._api.function_input import FunctionInputAPI
from cognite.powerops.client._generated.v1._api.function_input_query import FunctionInputQueryAPI
from cognite.powerops.client._generated.v1._api.function_output import FunctionOutputAPI
from cognite.powerops.client._generated.v1._api.function_output_alerts import FunctionOutputAlertsAPI
from cognite.powerops.client._generated.v1._api.function_output_query import FunctionOutputQueryAPI
from cognite.powerops.client._generated.v1._api.generator import GeneratorAPI
from cognite.powerops.client._generated.v1._api.generator_availability_time_series import GeneratorAvailabilityTimeSeriesAPI
from cognite.powerops.client._generated.v1._api.generator_efficiency_curve import GeneratorEfficiencyCurveAPI
from cognite.powerops.client._generated.v1._api.generator_efficiency_curve_query import GeneratorEfficiencyCurveQueryAPI
from cognite.powerops.client._generated.v1._api.generator_query import GeneratorQueryAPI
from cognite.powerops.client._generated.v1._api.generator_start_stop_cost_time_series import GeneratorStartStopCostTimeSeriesAPI
from cognite.powerops.client._generated.v1._api.generator_turbine_efficiency_curves import GeneratorTurbineEfficiencyCurvesAPI
from cognite.powerops.client._generated.v1._api.market_configuration import MarketConfigurationAPI
from cognite.powerops.client._generated.v1._api.market_configuration_query import MarketConfigurationQueryAPI
from cognite.powerops.client._generated.v1._api.multi_scenario_partial_bid_matrix_calculation_input import MultiScenarioPartialBidMatrixCalculationInputAPI
from cognite.powerops.client._generated.v1._api.multi_scenario_partial_bid_matrix_calculation_input_price_production import MultiScenarioPartialBidMatrixCalculationInputPriceProductionAPI
from cognite.powerops.client._generated.v1._api.multi_scenario_partial_bid_matrix_calculation_input_query import MultiScenarioPartialBidMatrixCalculationInputQueryAPI
from cognite.powerops.client._generated.v1._api.partial_bid_configuration import PartialBidConfigurationAPI
from cognite.powerops.client._generated.v1._api.partial_bid_configuration_query import PartialBidConfigurationQueryAPI
from cognite.powerops.client._generated.v1._api.partial_bid_matrix_calculation_input import PartialBidMatrixCalculationInputAPI
from cognite.powerops.client._generated.v1._api.partial_bid_matrix_calculation_input_query import PartialBidMatrixCalculationInputQueryAPI
from cognite.powerops.client._generated.v1._api.partial_bid_matrix_calculation_output import PartialBidMatrixCalculationOutputAPI
from cognite.powerops.client._generated.v1._api.partial_bid_matrix_calculation_output_alerts import PartialBidMatrixCalculationOutputAlertsAPI
from cognite.powerops.client._generated.v1._api.partial_bid_matrix_calculation_output_query import PartialBidMatrixCalculationOutputQueryAPI
from cognite.powerops.client._generated.v1._api.partial_bid_matrix_information import PartialBidMatrixInformationAPI
from cognite.powerops.client._generated.v1._api.partial_bid_matrix_information_alerts import PartialBidMatrixInformationAlertsAPI
from cognite.powerops.client._generated.v1._api.partial_bid_matrix_information_linked_time_series import PartialBidMatrixInformationLinkedTimeSeriesAPI
from cognite.powerops.client._generated.v1._api.partial_bid_matrix_information_query import PartialBidMatrixInformationQueryAPI
from cognite.powerops.client._generated.v1._api.partial_bid_matrix_information_underlying_bid_matrices import PartialBidMatrixInformationUnderlyingBidMatricesAPI
from cognite.powerops.client._generated.v1._api.partial_bid_matrix_information_with_scenarios import PartialBidMatrixInformationWithScenariosAPI
from cognite.powerops.client._generated.v1._api.partial_bid_matrix_information_with_scenarios_alerts import PartialBidMatrixInformationWithScenariosAlertsAPI
from cognite.powerops.client._generated.v1._api.partial_bid_matrix_information_with_scenarios_linked_time_series import PartialBidMatrixInformationWithScenariosLinkedTimeSeriesAPI
from cognite.powerops.client._generated.v1._api.partial_bid_matrix_information_with_scenarios_multi_scenario_input import PartialBidMatrixInformationWithScenariosMultiScenarioInputAPI
from cognite.powerops.client._generated.v1._api.partial_bid_matrix_information_with_scenarios_query import PartialBidMatrixInformationWithScenariosQueryAPI
from cognite.powerops.client._generated.v1._api.partial_bid_matrix_information_with_scenarios_underlying_bid_matrices import PartialBidMatrixInformationWithScenariosUnderlyingBidMatricesAPI
from cognite.powerops.client._generated.v1._api.plant import PlantAPI
from cognite.powerops.client._generated.v1._api.plant_information import PlantInformationAPI
from cognite.powerops.client._generated.v1._api.plant_information_feeding_fee_time_series import PlantInformationFeedingFeeTimeSeriesAPI
from cognite.powerops.client._generated.v1._api.plant_information_generators import PlantInformationGeneratorsAPI
from cognite.powerops.client._generated.v1._api.plant_information_head_direct_time_series import PlantInformationHeadDirectTimeSeriesAPI
from cognite.powerops.client._generated.v1._api.plant_information_inlet_level_time_series import PlantInformationInletLevelTimeSeriesAPI
from cognite.powerops.client._generated.v1._api.plant_information_outlet_level_time_series import PlantInformationOutletLevelTimeSeriesAPI
from cognite.powerops.client._generated.v1._api.plant_information_production_max_time_series import PlantInformationProductionMaxTimeSeriesAPI
from cognite.powerops.client._generated.v1._api.plant_information_production_min_time_series import PlantInformationProductionMinTimeSeriesAPI
from cognite.powerops.client._generated.v1._api.plant_information_query import PlantInformationQueryAPI
from cognite.powerops.client._generated.v1._api.plant_information_water_value_time_series import PlantInformationWaterValueTimeSeriesAPI
from cognite.powerops.client._generated.v1._api.plant_query import PlantQueryAPI
from cognite.powerops.client._generated.v1._api.plant_water_value_based import PlantWaterValueBasedAPI
from cognite.powerops.client._generated.v1._api.plant_water_value_based_feeding_fee_time_series import PlantWaterValueBasedFeedingFeeTimeSeriesAPI
from cognite.powerops.client._generated.v1._api.plant_water_value_based_generators import PlantWaterValueBasedGeneratorsAPI
from cognite.powerops.client._generated.v1._api.plant_water_value_based_head_direct_time_series import PlantWaterValueBasedHeadDirectTimeSeriesAPI
from cognite.powerops.client._generated.v1._api.plant_water_value_based_inlet_level_time_series import PlantWaterValueBasedInletLevelTimeSeriesAPI
from cognite.powerops.client._generated.v1._api.plant_water_value_based_outlet_level_time_series import PlantWaterValueBasedOutletLevelTimeSeriesAPI
from cognite.powerops.client._generated.v1._api.plant_water_value_based_production_max_time_series import PlantWaterValueBasedProductionMaxTimeSeriesAPI
from cognite.powerops.client._generated.v1._api.plant_water_value_based_production_min_time_series import PlantWaterValueBasedProductionMinTimeSeriesAPI
from cognite.powerops.client._generated.v1._api.plant_water_value_based_query import PlantWaterValueBasedQueryAPI
from cognite.powerops.client._generated.v1._api.plant_water_value_based_water_value_time_series import PlantWaterValueBasedWaterValueTimeSeriesAPI
from cognite.powerops.client._generated.v1._api.power_asset import PowerAssetAPI
from cognite.powerops.client._generated.v1._api.power_asset_query import PowerAssetQueryAPI
from cognite.powerops.client._generated.v1._api.price_area_afrr import PriceAreaAFRRAPI
from cognite.powerops.client._generated.v1._api.price_area_afrr_activation_price_down import PriceAreaAFRRActivationPriceDownAPI
from cognite.powerops.client._generated.v1._api.price_area_afrr_activation_price_up import PriceAreaAFRRActivationPriceUpAPI
from cognite.powerops.client._generated.v1._api.price_area_afrr_capacity_price_down import PriceAreaAFRRCapacityPriceDownAPI
from cognite.powerops.client._generated.v1._api.price_area_afrr_capacity_price_up import PriceAreaAFRRCapacityPriceUpAPI
from cognite.powerops.client._generated.v1._api.price_area_afrr_own_capacity_allocation_down import PriceAreaAFRROwnCapacityAllocationDownAPI
from cognite.powerops.client._generated.v1._api.price_area_afrr_own_capacity_allocation_up import PriceAreaAFRROwnCapacityAllocationUpAPI
from cognite.powerops.client._generated.v1._api.price_area_afrr_query import PriceAreaAFRRQueryAPI
from cognite.powerops.client._generated.v1._api.price_area_afrr_relative_activation import PriceAreaAFRRRelativeActivationAPI
from cognite.powerops.client._generated.v1._api.price_area_afrr_total_capacity_allocation_down import PriceAreaAFRRTotalCapacityAllocationDownAPI
from cognite.powerops.client._generated.v1._api.price_area_afrr_total_capacity_allocation_up import PriceAreaAFRRTotalCapacityAllocationUpAPI
from cognite.powerops.client._generated.v1._api.price_area import PriceAreaAPI
from cognite.powerops.client._generated.v1._api.price_area_day_ahead import PriceAreaDayAheadAPI
from cognite.powerops.client._generated.v1._api.price_area_day_ahead_main_price_scenario import PriceAreaDayAheadMainPriceScenarioAPI
from cognite.powerops.client._generated.v1._api.price_area_day_ahead_price_scenarios import PriceAreaDayAheadPriceScenariosAPI
from cognite.powerops.client._generated.v1._api.price_area_day_ahead_query import PriceAreaDayAheadQueryAPI
from cognite.powerops.client._generated.v1._api.price_area_information import PriceAreaInformationAPI
from cognite.powerops.client._generated.v1._api.price_area_information_activation_price_down import PriceAreaInformationActivationPriceDownAPI
from cognite.powerops.client._generated.v1._api.price_area_information_activation_price_up import PriceAreaInformationActivationPriceUpAPI
from cognite.powerops.client._generated.v1._api.price_area_information_capacity_price_down import PriceAreaInformationCapacityPriceDownAPI
from cognite.powerops.client._generated.v1._api.price_area_information_capacity_price_up import PriceAreaInformationCapacityPriceUpAPI
from cognite.powerops.client._generated.v1._api.price_area_information_main_price_scenario import PriceAreaInformationMainPriceScenarioAPI
from cognite.powerops.client._generated.v1._api.price_area_information_own_capacity_allocation_down import PriceAreaInformationOwnCapacityAllocationDownAPI
from cognite.powerops.client._generated.v1._api.price_area_information_own_capacity_allocation_up import PriceAreaInformationOwnCapacityAllocationUpAPI
from cognite.powerops.client._generated.v1._api.price_area_information_price_scenarios import PriceAreaInformationPriceScenariosAPI
from cognite.powerops.client._generated.v1._api.price_area_information_query import PriceAreaInformationQueryAPI
from cognite.powerops.client._generated.v1._api.price_area_information_relative_activation import PriceAreaInformationRelativeActivationAPI
from cognite.powerops.client._generated.v1._api.price_area_information_total_capacity_allocation_down import PriceAreaInformationTotalCapacityAllocationDownAPI
from cognite.powerops.client._generated.v1._api.price_area_information_total_capacity_allocation_up import PriceAreaInformationTotalCapacityAllocationUpAPI
from cognite.powerops.client._generated.v1._api.price_area_query import PriceAreaQueryAPI
from cognite.powerops.client._generated.v1._api.price_production import PriceProductionAPI
from cognite.powerops.client._generated.v1._api.price_production_price import PriceProductionPriceAPI
from cognite.powerops.client._generated.v1._api.price_production_production import PriceProductionProductionAPI
from cognite.powerops.client._generated.v1._api.price_production_query import PriceProductionQueryAPI
from cognite.powerops.client._generated.v1._api.shop_attribute_mapping import ShopAttributeMappingAPI
from cognite.powerops.client._generated.v1._api.shop_attribute_mapping_query import ShopAttributeMappingQueryAPI
from cognite.powerops.client._generated.v1._api.shop_attribute_mapping_time_series import ShopAttributeMappingTimeSeriesAPI
from cognite.powerops.client._generated.v1._api.shop_based_partial_bid_configuration import ShopBasedPartialBidConfigurationAPI
from cognite.powerops.client._generated.v1._api.shop_based_partial_bid_configuration_query import ShopBasedPartialBidConfigurationQueryAPI
from cognite.powerops.client._generated.v1._api.shop_case import ShopCaseAPI
from cognite.powerops.client._generated.v1._api.shop_case_query import ShopCaseQueryAPI
from cognite.powerops.client._generated.v1._api.shop_case_shop_files import ShopCaseShopFilesAPI
from cognite.powerops.client._generated.v1._api.shop_commands import ShopCommandsAPI
from cognite.powerops.client._generated.v1._api.shop_commands_query import ShopCommandsQueryAPI
from cognite.powerops.client._generated.v1._api.shop_file import ShopFileAPI
from cognite.powerops.client._generated.v1._api.shop_file_query import ShopFileQueryAPI
from cognite.powerops.client._generated.v1._api.shop_model import ShopModelAPI
from cognite.powerops.client._generated.v1._api.shop_model_base_attribute_mappings import ShopModelBaseAttributeMappingsAPI
from cognite.powerops.client._generated.v1._api.shop_model_cog_shop_files_config import ShopModelCogShopFilesConfigAPI
from cognite.powerops.client._generated.v1._api.shop_model_query import ShopModelQueryAPI
from cognite.powerops.client._generated.v1._api.shop_model_with_assets import ShopModelWithAssetsAPI
from cognite.powerops.client._generated.v1._api.shop_model_with_assets_power_assets import ShopModelWithAssetsPowerAssetsAPI
from cognite.powerops.client._generated.v1._api.shop_model_with_assets_production_obligations import ShopModelWithAssetsProductionObligationsAPI
from cognite.powerops.client._generated.v1._api.shop_model_with_assets_query import ShopModelWithAssetsQueryAPI
from cognite.powerops.client._generated.v1._api.shop_output_time_series_definition import ShopOutputTimeSeriesDefinitionAPI
from cognite.powerops.client._generated.v1._api.shop_output_time_series_definition_query import ShopOutputTimeSeriesDefinitionQueryAPI
from cognite.powerops.client._generated.v1._api.shop_penalty_report import ShopPenaltyReportAPI
from cognite.powerops.client._generated.v1._api.shop_penalty_report_query import ShopPenaltyReportQueryAPI
from cognite.powerops.client._generated.v1._api.shop_preprocessor_input import ShopPreprocessorInputAPI
from cognite.powerops.client._generated.v1._api.shop_preprocessor_input_query import ShopPreprocessorInputQueryAPI
from cognite.powerops.client._generated.v1._api.shop_preprocessor_output import ShopPreprocessorOutputAPI
from cognite.powerops.client._generated.v1._api.shop_preprocessor_output_alerts import ShopPreprocessorOutputAlertsAPI
from cognite.powerops.client._generated.v1._api.shop_preprocessor_output_query import ShopPreprocessorOutputQueryAPI
from cognite.powerops.client._generated.v1._api.shop_result import ShopResultAPI
from cognite.powerops.client._generated.v1._api.shop_result_alerts import ShopResultAlertsAPI
from cognite.powerops.client._generated.v1._api.shop_result_output_time_series import ShopResultOutputTimeSeriesAPI
from cognite.powerops.client._generated.v1._api.shop_result_query import ShopResultQueryAPI
from cognite.powerops.client._generated.v1._api.shop_scenario import ShopScenarioAPI
from cognite.powerops.client._generated.v1._api.shop_scenario_attribute_mappings_override import ShopScenarioAttributeMappingsOverrideAPI
from cognite.powerops.client._generated.v1._api.shop_scenario_output_definition import ShopScenarioOutputDefinitionAPI
from cognite.powerops.client._generated.v1._api.shop_scenario_query import ShopScenarioQueryAPI
from cognite.powerops.client._generated.v1._api.shop_scenario_set import ShopScenarioSetAPI
from cognite.powerops.client._generated.v1._api.shop_scenario_set_query import ShopScenarioSetQueryAPI
from cognite.powerops.client._generated.v1._api.shop_scenario_set_scenarios import ShopScenarioSetScenariosAPI
from cognite.powerops.client._generated.v1._api.shop_time_resolution import ShopTimeResolutionAPI
from cognite.powerops.client._generated.v1._api.shop_time_resolution_query import ShopTimeResolutionQueryAPI
from cognite.powerops.client._generated.v1._api.shop_time_series import ShopTimeSeriesAPI
from cognite.powerops.client._generated.v1._api.shop_time_series_query import ShopTimeSeriesQueryAPI
from cognite.powerops.client._generated.v1._api.shop_time_series_time_series import ShopTimeSeriesTimeSeriesAPI
from cognite.powerops.client._generated.v1._api.shop_trigger_input import ShopTriggerInputAPI
from cognite.powerops.client._generated.v1._api.shop_trigger_input_query import ShopTriggerInputQueryAPI
from cognite.powerops.client._generated.v1._api.shop_trigger_output import ShopTriggerOutputAPI
from cognite.powerops.client._generated.v1._api.shop_trigger_output_alerts import ShopTriggerOutputAlertsAPI
from cognite.powerops.client._generated.v1._api.shop_trigger_output_query import ShopTriggerOutputQueryAPI
from cognite.powerops.client._generated.v1._api.task_dispatcher_input import TaskDispatcherInputAPI
from cognite.powerops.client._generated.v1._api.task_dispatcher_input_query import TaskDispatcherInputQueryAPI
from cognite.powerops.client._generated.v1._api.task_dispatcher_output import TaskDispatcherOutputAPI
from cognite.powerops.client._generated.v1._api.task_dispatcher_output_alerts import TaskDispatcherOutputAlertsAPI
from cognite.powerops.client._generated.v1._api.task_dispatcher_output_process_sub_tasks import TaskDispatcherOutputProcessSubTasksAPI
from cognite.powerops.client._generated.v1._api.task_dispatcher_output_query import TaskDispatcherOutputQueryAPI
from cognite.powerops.client._generated.v1._api.total_bid_matrix_calculation_input import TotalBidMatrixCalculationInputAPI
from cognite.powerops.client._generated.v1._api.total_bid_matrix_calculation_input_partial_bid_matrices import TotalBidMatrixCalculationInputPartialBidMatricesAPI
from cognite.powerops.client._generated.v1._api.total_bid_matrix_calculation_input_query import TotalBidMatrixCalculationInputQueryAPI
from cognite.powerops.client._generated.v1._api.total_bid_matrix_calculation_output import TotalBidMatrixCalculationOutputAPI
from cognite.powerops.client._generated.v1._api.total_bid_matrix_calculation_output_alerts import TotalBidMatrixCalculationOutputAlertsAPI
from cognite.powerops.client._generated.v1._api.total_bid_matrix_calculation_output_query import TotalBidMatrixCalculationOutputQueryAPI
from cognite.powerops.client._generated.v1._api.turbine_efficiency_curve import TurbineEfficiencyCurveAPI
from cognite.powerops.client._generated.v1._api.turbine_efficiency_curve_query import TurbineEfficiencyCurveQueryAPI
from cognite.powerops.client._generated.v1._api.water_value_based_partial_bid_configuration import WaterValueBasedPartialBidConfigurationAPI
from cognite.powerops.client._generated.v1._api.water_value_based_partial_bid_configuration_query import WaterValueBasedPartialBidConfigurationQueryAPI
from cognite.powerops.client._generated.v1._api.water_value_based_partial_bid_matrix_calculation_input import WaterValueBasedPartialBidMatrixCalculationInputAPI
from cognite.powerops.client._generated.v1._api.water_value_based_partial_bid_matrix_calculation_input_query import WaterValueBasedPartialBidMatrixCalculationInputQueryAPI
from cognite.powerops.client._generated.v1._api.watercourse import WatercourseAPI
from cognite.powerops.client._generated.v1._api.watercourse_query import WatercourseQueryAPI

__all__ = [
    "AlertAPI",
    "AlertQueryAPI",
    "BenchmarkingCalculationInputAPI",
    "BenchmarkingCalculationInputQueryAPI",
    "BenchmarkingCalculationInputShopResultsAPI",
    "BenchmarkingCalculationOutputAPI",
    "BenchmarkingCalculationOutputAlertsAPI",
    "BenchmarkingCalculationOutputBenchmarkingResultsAPI",
    "BenchmarkingCalculationOutputQueryAPI",
    "BenchmarkingConfigurationDayAheadAPI",
    "BenchmarkingConfigurationDayAheadAssetsPerShopModelAPI",
    "BenchmarkingConfigurationDayAheadBidConfigurationsAPI",
    "BenchmarkingConfigurationDayAheadQueryAPI",
    "BenchmarkingProductionObligationDayAheadAPI",
    "BenchmarkingProductionObligationDayAheadQueryAPI",
    "BenchmarkingProductionObligationDayAheadTimeSeriesAPI",
    "BenchmarkingResultDayAheadAPI",
    "BenchmarkingResultDayAheadAlertsAPI",
    "BenchmarkingResultDayAheadQueryAPI",
    "BenchmarkingShopCaseAPI",
    "BenchmarkingShopCaseQueryAPI",
    "BenchmarkingShopCaseShopFilesAPI",
    "BenchmarkingTaskDispatcherInputDayAheadAPI",
    "BenchmarkingTaskDispatcherInputDayAheadQueryAPI",
    "BenchmarkingTaskDispatcherOutputDayAheadAPI",
    "BenchmarkingTaskDispatcherOutputDayAheadAlertsAPI",
    "BenchmarkingTaskDispatcherOutputDayAheadBenchmarkingSubTasksAPI",
    "BenchmarkingTaskDispatcherOutputDayAheadQueryAPI",
    "BidConfigurationDayAheadAPI",
    "BidConfigurationDayAheadPartialsAPI",
    "BidConfigurationDayAheadQueryAPI",
    "BidDocumentAFRRAPI",
    "BidDocumentAFRRAlertsAPI",
    "BidDocumentAFRRBidsAPI",
    "BidDocumentAFRRQueryAPI",
    "BidDocumentAPI",
    "BidDocumentAlertsAPI",
    "BidDocumentDayAheadAPI",
    "BidDocumentDayAheadAlertsAPI",
    "BidDocumentDayAheadPartialsAPI",
    "BidDocumentDayAheadQueryAPI",
    "BidDocumentQueryAPI",
    "BidMatrixAPI",
    "BidMatrixInformationAPI",
    "BidMatrixInformationAlertsAPI",
    "BidMatrixInformationLinkedTimeSeriesAPI",
    "BidMatrixInformationQueryAPI",
    "BidMatrixInformationUnderlyingBidMatricesAPI",
    "BidMatrixQueryAPI",
    "BidRowAPI",
    "BidRowAlertsAPI",
    "BidRowQueryAPI",
    "DateSpecificationAPI",
    "DateSpecificationQueryAPI",
    "FunctionInputAPI",
    "FunctionInputQueryAPI",
    "FunctionOutputAPI",
    "FunctionOutputAlertsAPI",
    "FunctionOutputQueryAPI",
    "GeneratorAPI",
    "GeneratorAvailabilityTimeSeriesAPI",
    "GeneratorEfficiencyCurveAPI",
    "GeneratorEfficiencyCurveQueryAPI",
    "GeneratorQueryAPI",
    "GeneratorStartStopCostTimeSeriesAPI",
    "GeneratorTurbineEfficiencyCurvesAPI",
    "MarketConfigurationAPI",
    "MarketConfigurationQueryAPI",
    "MultiScenarioPartialBidMatrixCalculationInputAPI",
    "MultiScenarioPartialBidMatrixCalculationInputPriceProductionAPI",
    "MultiScenarioPartialBidMatrixCalculationInputQueryAPI",
    "PartialBidConfigurationAPI",
    "PartialBidConfigurationQueryAPI",
    "PartialBidMatrixCalculationInputAPI",
    "PartialBidMatrixCalculationInputQueryAPI",
    "PartialBidMatrixCalculationOutputAPI",
    "PartialBidMatrixCalculationOutputAlertsAPI",
    "PartialBidMatrixCalculationOutputQueryAPI",
    "PartialBidMatrixInformationAPI",
    "PartialBidMatrixInformationAlertsAPI",
    "PartialBidMatrixInformationLinkedTimeSeriesAPI",
    "PartialBidMatrixInformationQueryAPI",
    "PartialBidMatrixInformationUnderlyingBidMatricesAPI",
    "PartialBidMatrixInformationWithScenariosAPI",
    "PartialBidMatrixInformationWithScenariosAlertsAPI",
    "PartialBidMatrixInformationWithScenariosLinkedTimeSeriesAPI",
    "PartialBidMatrixInformationWithScenariosMultiScenarioInputAPI",
    "PartialBidMatrixInformationWithScenariosQueryAPI",
    "PartialBidMatrixInformationWithScenariosUnderlyingBidMatricesAPI",
    "PlantAPI",
    "PlantInformationAPI",
    "PlantInformationFeedingFeeTimeSeriesAPI",
    "PlantInformationGeneratorsAPI",
    "PlantInformationHeadDirectTimeSeriesAPI",
    "PlantInformationInletLevelTimeSeriesAPI",
    "PlantInformationOutletLevelTimeSeriesAPI",
    "PlantInformationProductionMaxTimeSeriesAPI",
    "PlantInformationProductionMinTimeSeriesAPI",
    "PlantInformationQueryAPI",
    "PlantInformationWaterValueTimeSeriesAPI",
    "PlantQueryAPI",
    "PlantWaterValueBasedAPI",
    "PlantWaterValueBasedFeedingFeeTimeSeriesAPI",
    "PlantWaterValueBasedGeneratorsAPI",
    "PlantWaterValueBasedHeadDirectTimeSeriesAPI",
    "PlantWaterValueBasedInletLevelTimeSeriesAPI",
    "PlantWaterValueBasedOutletLevelTimeSeriesAPI",
    "PlantWaterValueBasedProductionMaxTimeSeriesAPI",
    "PlantWaterValueBasedProductionMinTimeSeriesAPI",
    "PlantWaterValueBasedQueryAPI",
    "PlantWaterValueBasedWaterValueTimeSeriesAPI",
    "PowerAssetAPI",
    "PowerAssetQueryAPI",
    "PriceAreaAFRRAPI",
    "PriceAreaAFRRActivationPriceDownAPI",
    "PriceAreaAFRRActivationPriceUpAPI",
    "PriceAreaAFRRCapacityPriceDownAPI",
    "PriceAreaAFRRCapacityPriceUpAPI",
    "PriceAreaAFRROwnCapacityAllocationDownAPI",
    "PriceAreaAFRROwnCapacityAllocationUpAPI",
    "PriceAreaAFRRQueryAPI",
    "PriceAreaAFRRRelativeActivationAPI",
    "PriceAreaAFRRTotalCapacityAllocationDownAPI",
    "PriceAreaAFRRTotalCapacityAllocationUpAPI",
    "PriceAreaAPI",
    "PriceAreaDayAheadAPI",
    "PriceAreaDayAheadMainPriceScenarioAPI",
    "PriceAreaDayAheadPriceScenariosAPI",
    "PriceAreaDayAheadQueryAPI",
    "PriceAreaInformationAPI",
    "PriceAreaInformationActivationPriceDownAPI",
    "PriceAreaInformationActivationPriceUpAPI",
    "PriceAreaInformationCapacityPriceDownAPI",
    "PriceAreaInformationCapacityPriceUpAPI",
    "PriceAreaInformationMainPriceScenarioAPI",
    "PriceAreaInformationOwnCapacityAllocationDownAPI",
    "PriceAreaInformationOwnCapacityAllocationUpAPI",
    "PriceAreaInformationPriceScenariosAPI",
    "PriceAreaInformationQueryAPI",
    "PriceAreaInformationRelativeActivationAPI",
    "PriceAreaInformationTotalCapacityAllocationDownAPI",
    "PriceAreaInformationTotalCapacityAllocationUpAPI",
    "PriceAreaQueryAPI",
    "PriceProductionAPI",
    "PriceProductionPriceAPI",
    "PriceProductionProductionAPI",
    "PriceProductionQueryAPI",
    "ShopAttributeMappingAPI",
    "ShopAttributeMappingQueryAPI",
    "ShopAttributeMappingTimeSeriesAPI",
    "ShopBasedPartialBidConfigurationAPI",
    "ShopBasedPartialBidConfigurationQueryAPI",
    "ShopCaseAPI",
    "ShopCaseQueryAPI",
    "ShopCaseShopFilesAPI",
    "ShopCommandsAPI",
    "ShopCommandsQueryAPI",
    "ShopFileAPI",
    "ShopFileQueryAPI",
    "ShopModelAPI",
    "ShopModelBaseAttributeMappingsAPI",
    "ShopModelCogShopFilesConfigAPI",
    "ShopModelQueryAPI",
    "ShopModelWithAssetsAPI",
    "ShopModelWithAssetsPowerAssetsAPI",
    "ShopModelWithAssetsProductionObligationsAPI",
    "ShopModelWithAssetsQueryAPI",
    "ShopOutputTimeSeriesDefinitionAPI",
    "ShopOutputTimeSeriesDefinitionQueryAPI",
    "ShopPenaltyReportAPI",
    "ShopPenaltyReportQueryAPI",
    "ShopPreprocessorInputAPI",
    "ShopPreprocessorInputQueryAPI",
    "ShopPreprocessorOutputAPI",
    "ShopPreprocessorOutputAlertsAPI",
    "ShopPreprocessorOutputQueryAPI",
    "ShopResultAPI",
    "ShopResultAlertsAPI",
    "ShopResultOutputTimeSeriesAPI",
    "ShopResultQueryAPI",
    "ShopScenarioAPI",
    "ShopScenarioAttributeMappingsOverrideAPI",
    "ShopScenarioOutputDefinitionAPI",
    "ShopScenarioQueryAPI",
    "ShopScenarioSetAPI",
    "ShopScenarioSetQueryAPI",
    "ShopScenarioSetScenariosAPI",
    "ShopTimeResolutionAPI",
    "ShopTimeResolutionQueryAPI",
    "ShopTimeSeriesAPI",
    "ShopTimeSeriesQueryAPI",
    "ShopTimeSeriesTimeSeriesAPI",
    "ShopTriggerInputAPI",
    "ShopTriggerInputQueryAPI",
    "ShopTriggerOutputAPI",
    "ShopTriggerOutputAlertsAPI",
    "ShopTriggerOutputQueryAPI",
    "TaskDispatcherInputAPI",
    "TaskDispatcherInputQueryAPI",
    "TaskDispatcherOutputAPI",
    "TaskDispatcherOutputAlertsAPI",
    "TaskDispatcherOutputProcessSubTasksAPI",
    "TaskDispatcherOutputQueryAPI",
    "TotalBidMatrixCalculationInputAPI",
    "TotalBidMatrixCalculationInputPartialBidMatricesAPI",
    "TotalBidMatrixCalculationInputQueryAPI",
    "TotalBidMatrixCalculationOutputAPI",
    "TotalBidMatrixCalculationOutputAlertsAPI",
    "TotalBidMatrixCalculationOutputQueryAPI",
    "TurbineEfficiencyCurveAPI",
    "TurbineEfficiencyCurveQueryAPI",
    "WaterValueBasedPartialBidConfigurationAPI",
    "WaterValueBasedPartialBidConfigurationQueryAPI",
    "WaterValueBasedPartialBidMatrixCalculationInputAPI",
    "WaterValueBasedPartialBidMatrixCalculationInputQueryAPI",
    "WatercourseAPI",
    "WatercourseQueryAPI",
]



from __future__ import annotations

import warnings
from collections.abc import Sequence
from typing import TYPE_CHECKING, Any, ClassVar, Literal,  no_type_check, Optional, Union

from cognite.client import data_modeling as dm, CogniteClient
from pydantic import Field
from pydantic import field_validator, model_validator

from cognite.powerops.client._generated.v1.data_classes._core import (
    DEFAULT_INSTANCE_SPACE,
    DEFAULT_QUERY_LIMIT,
    DataRecord,
    DataRecordGraphQL,
    DataRecordWrite,
    DomainModel,
    DomainModelWrite,
    DomainModelWriteList,
    DomainModelList,
    DomainRelation,
    DomainRelationWrite,
    GraphQLCore,
    ResourcesWrite,
    T_DomainModelList,
    as_direct_relation_reference,
    as_instance_dict_id,
    as_node_id,
    as_pygen_node_id,
    are_nodes_equal,
    is_tuple_id,
    select_best_node,
    QueryCore,
    NodeQueryCore,
    StringFilter,

)

if TYPE_CHECKING:
    from cognite.powerops.client._generated.v1.data_classes._shop_attribute_mapping import ShopAttributeMapping, ShopAttributeMappingList, ShopAttributeMappingGraphQL, ShopAttributeMappingWrite, ShopAttributeMappingWriteList
    from cognite.powerops.client._generated.v1.data_classes._shop_commands import ShopCommands, ShopCommandsList, ShopCommandsGraphQL, ShopCommandsWrite, ShopCommandsWriteList
    from cognite.powerops.client._generated.v1.data_classes._shop_model import ShopModel, ShopModelList, ShopModelGraphQL, ShopModelWrite, ShopModelWriteList
    from cognite.powerops.client._generated.v1.data_classes._shop_output_time_series_definition import ShopOutputTimeSeriesDefinition, ShopOutputTimeSeriesDefinitionList, ShopOutputTimeSeriesDefinitionGraphQL, ShopOutputTimeSeriesDefinitionWrite, ShopOutputTimeSeriesDefinitionWriteList
    from cognite.powerops.client._generated.v1.data_classes._shop_time_resolution import ShopTimeResolution, ShopTimeResolutionList, ShopTimeResolutionGraphQL, ShopTimeResolutionWrite, ShopTimeResolutionWriteList


__all__ = [
    "ShopScenario",
    "ShopScenarioWrite",
    "ShopScenarioApply",
    "ShopScenarioList",
    "ShopScenarioWriteList",
    "ShopScenarioApplyList",
    "ShopScenarioFields",
    "ShopScenarioTextFields",
    "ShopScenarioGraphQL",
]


ShopScenarioTextFields = Literal["external_id", "name", "source"]
ShopScenarioFields = Literal["external_id", "name", "source"]

_SHOPSCENARIO_PROPERTIES_BY_FIELD = {
    "external_id": "externalId",
    "name": "name",
    "source": "source",
}

class ShopScenarioGraphQL(GraphQLCore, protected_namespaces=()):
    """This represents the reading version of shop scenario, used
    when data is retrieved from CDF using GraphQL.

    It is used when retrieving data from CDF using GraphQL.

    Args:
        space: The space where the node is located.
        external_id: The external id of the shop scenario.
        data_record: The data record of the shop scenario node.
        name: The name of the scenario to run
        model: The model template to use when running the scenario
        commands: The commands to run
        source: The source of the scenario
        time_resolution: The time resolutions to use within SHOP.
        output_definition: An array of output definitions for the time series
        attribute_mappings_override: An array of base mappings to override in shop model file
    """
    view_id: ClassVar[dm.ViewId] = dm.ViewId("power_ops_core", "ShopScenario", "1")
    name: Optional[str] = None
    model: Optional[ShopModelGraphQL] = Field(default=None, repr=False)
    commands: Optional[ShopCommandsGraphQL] = Field(default=None, repr=False)
    source: Optional[str] = None
    time_resolution: Optional[ShopTimeResolutionGraphQL] = Field(default=None, repr=False, alias="timeResolution")
    output_definition: Optional[list[ShopOutputTimeSeriesDefinitionGraphQL]] = Field(default=None, repr=False, alias="outputDefinition")
    attribute_mappings_override: Optional[list[ShopAttributeMappingGraphQL]] = Field(default=None, repr=False, alias="attributeMappingsOverride")

    @model_validator(mode="before")
    def parse_data_record(cls, values: Any) -> Any:
        if not isinstance(values, dict):
            return values
        if "lastUpdatedTime" in values or "createdTime" in values:
            values["dataRecord"] = DataRecordGraphQL(
                created_time=values.pop("createdTime", None),
                last_updated_time=values.pop("lastUpdatedTime", None),
            )
        return values
    @field_validator("model", "commands", "time_resolution", "output_definition", "attribute_mappings_override", mode="before")
    def parse_graphql(cls, value: Any) -> Any:
        if not isinstance(value, dict):
            return value
        if "items" in value:
            return value["items"]
        return value

    # We do the ignore argument type as we let pydantic handle the type checking
    @no_type_check
    def as_read(self) -> ShopScenario:
        """Convert this GraphQL format of shop scenario to the reading format."""
        if self.data_record is None:
            raise ValueError("This object cannot be converted to a read format because it lacks a data record.")
        return ShopScenario(
            space=self.space,
            external_id=self.external_id,
            data_record=DataRecord(
                version=0,
                last_updated_time=self.data_record.last_updated_time,
                created_time=self.data_record.created_time,
            ),
            name=self.name,
            model=self.model.as_read()
if isinstance(self.model, GraphQLCore)
else self.model,
            commands=self.commands.as_read()
if isinstance(self.commands, GraphQLCore)
else self.commands,
            source=self.source,
            time_resolution=self.time_resolution.as_read()
if isinstance(self.time_resolution, GraphQLCore)
else self.time_resolution,
            output_definition=[output_definition.as_read() for output_definition in self.output_definition or []],
            attribute_mappings_override=[attribute_mappings_override.as_read() for attribute_mappings_override in self.attribute_mappings_override or []],
        )


    # We do the ignore argument type as we let pydantic handle the type checking
    @no_type_check
    def as_write(self) -> ShopScenarioWrite:
        """Convert this GraphQL format of shop scenario to the writing format."""
        return ShopScenarioWrite(
            space=self.space,
            external_id=self.external_id,
            data_record=DataRecordWrite(existing_version=0),
            name=self.name,
            model=self.model.as_write()
if isinstance(self.model, GraphQLCore)
else self.model,
            commands=self.commands.as_write()
if isinstance(self.commands, GraphQLCore)
else self.commands,
            source=self.source,
            time_resolution=self.time_resolution.as_write()
if isinstance(self.time_resolution, GraphQLCore)
else self.time_resolution,
            output_definition=[output_definition.as_write() for output_definition in self.output_definition or []],
            attribute_mappings_override=[attribute_mappings_override.as_write() for attribute_mappings_override in self.attribute_mappings_override or []],
        )


class ShopScenario(DomainModel, protected_namespaces=()):
    """This represents the reading version of shop scenario.

    It is used to when data is retrieved from CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the shop scenario.
        data_record: The data record of the shop scenario node.
        name: The name of the scenario to run
        model: The model template to use when running the scenario
        commands: The commands to run
        source: The source of the scenario
        time_resolution: The time resolutions to use within SHOP.
        output_definition: An array of output definitions for the time series
        attribute_mappings_override: An array of base mappings to override in shop model file
    """
    _view_id: ClassVar[dm.ViewId] = dm.ViewId("power_ops_core", "ShopScenario", "1")

    space: str = DEFAULT_INSTANCE_SPACE
    node_type: Union[dm.DirectRelationReference, None] = None
    name: str
    model: Union[ShopModel, str, dm.NodeId, None] = Field(default=None, repr=False)
    commands: Union[ShopCommands, str, dm.NodeId, None] = Field(default=None, repr=False)
    source: Optional[str] = None
    time_resolution: Union[ShopTimeResolution, str, dm.NodeId, None] = Field(default=None, repr=False, alias="timeResolution")
    output_definition: Optional[list[Union[ShopOutputTimeSeriesDefinition, str, dm.NodeId]]] = Field(default=None, repr=False, alias="outputDefinition")
    attribute_mappings_override: Optional[list[Union[ShopAttributeMapping, str, dm.NodeId]]] = Field(default=None, repr=False, alias="attributeMappingsOverride")

    def as_write(self) -> ShopScenarioWrite:
        """Convert this read version of shop scenario to the writing version."""
        return ShopScenarioWrite(
            space=self.space,
            external_id=self.external_id,
            data_record=DataRecordWrite(existing_version=self.data_record.version),
            name=self.name,
            model=self.model.as_write()
if isinstance(self.model, DomainModel)
else self.model,
            commands=self.commands.as_write()
if isinstance(self.commands, DomainModel)
else self.commands,
            source=self.source,
            time_resolution=self.time_resolution.as_write()
if isinstance(self.time_resolution, DomainModel)
else self.time_resolution,
            output_definition=[output_definition.as_write() if isinstance(output_definition, DomainModel) else output_definition for output_definition in self.output_definition or []],
            attribute_mappings_override=[attribute_mappings_override.as_write() if isinstance(attribute_mappings_override, DomainModel) else attribute_mappings_override for attribute_mappings_override in self.attribute_mappings_override or []],
        )

    def as_apply(self) -> ShopScenarioWrite:
        """Convert this read version of shop scenario to the writing version."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()

    @classmethod
    def _update_connections(
        cls,
        instances: dict[dm.NodeId | str, ShopScenario],  # type: ignore[override]
        nodes_by_id: dict[dm.NodeId | str, DomainModel],
        edges_by_source_node: dict[dm.NodeId, list[dm.Edge | DomainRelation]],
    ) -> None:
        from ._shop_attribute_mapping import ShopAttributeMapping
        from ._shop_commands import ShopCommands
        from ._shop_model import ShopModel
        from ._shop_output_time_series_definition import ShopOutputTimeSeriesDefinition
        from ._shop_time_resolution import ShopTimeResolution

        for instance in instances.values():
            if isinstance(instance.model, (dm.NodeId, str)) and (model := nodes_by_id.get(instance.model)) and isinstance(
                    model, ShopModel
            ):
                instance.model = model
            if isinstance(instance.commands, (dm.NodeId, str)) and (commands := nodes_by_id.get(instance.commands)) and isinstance(
                    commands, ShopCommands
            ):
                instance.commands = commands
            if isinstance(instance.time_resolution, (dm.NodeId, str)) and (time_resolution := nodes_by_id.get(instance.time_resolution)) and isinstance(
                    time_resolution, ShopTimeResolution
            ):
                instance.time_resolution = time_resolution
            if edges := edges_by_source_node.get(instance.as_id()):
                output_definition: list[ShopOutputTimeSeriesDefinition | str | dm.NodeId] = []
                attribute_mappings_override: list[ShopAttributeMapping | str | dm.NodeId] = []
                for edge in edges:
                    value: DomainModel | DomainRelation | str | dm.NodeId
                    if isinstance(edge, DomainRelation):
                        value = edge
                    else:
                        other_end: dm.DirectRelationReference = (
                            edge.end_node
                            if edge.start_node.space == instance.space
                            and edge.start_node.external_id == instance.external_id
                            else edge.start_node
                        )
                        destination: dm.NodeId | str = (
                            as_node_id(other_end)
                            if other_end.space != DEFAULT_INSTANCE_SPACE
                            else other_end.external_id
                        )
                        if destination in nodes_by_id:
                            value = nodes_by_id[destination]
                        else:
                            value = destination
                    edge_type = edge.edge_type if isinstance(edge, DomainRelation) else edge.type

                    if edge_type == dm.DirectRelationReference("power_ops_types", "ShopOutputTimeSeriesDefinition") and isinstance(
                        value, (ShopOutputTimeSeriesDefinition, str, dm.NodeId)
                    ):
                        output_definition.append(value)
                    if edge_type == dm.DirectRelationReference("power_ops_types", "ShopAttributeMapping") and isinstance(
                        value, (ShopAttributeMapping, str, dm.NodeId)
                    ):
                        attribute_mappings_override.append(value)

                instance.output_definition = output_definition or None
                instance.attribute_mappings_override = attribute_mappings_override or None




class ShopScenarioWrite(DomainModelWrite, protected_namespaces=()):
    """This represents the writing version of shop scenario.

    It is used to when data is sent to CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the shop scenario.
        data_record: The data record of the shop scenario node.
        name: The name of the scenario to run
        model: The model template to use when running the scenario
        commands: The commands to run
        source: The source of the scenario
        time_resolution: The time resolutions to use within SHOP.
        output_definition: An array of output definitions for the time series
        attribute_mappings_override: An array of base mappings to override in shop model file
    """
    _view_id: ClassVar[dm.ViewId] = dm.ViewId("power_ops_core", "ShopScenario", "1")

    space: str = DEFAULT_INSTANCE_SPACE
    node_type: Union[dm.DirectRelationReference, dm.NodeId, tuple[str, str], None] = None
    name: str
    model: Union[ShopModelWrite, str, dm.NodeId, None] = Field(default=None, repr=False)
    commands: Union[ShopCommandsWrite, str, dm.NodeId, None] = Field(default=None, repr=False)
    source: Optional[str] = None
    time_resolution: Union[ShopTimeResolutionWrite, str, dm.NodeId, None] = Field(default=None, repr=False, alias="timeResolution")
    output_definition: Optional[list[Union[ShopOutputTimeSeriesDefinitionWrite, str, dm.NodeId]]] = Field(default=None, repr=False, alias="outputDefinition")
    attribute_mappings_override: Optional[list[Union[ShopAttributeMappingWrite, str, dm.NodeId]]] = Field(default=None, repr=False, alias="attributeMappingsOverride")

    @field_validator("model", "commands", "time_resolution", "output_definition", "attribute_mappings_override", mode="before")
    def as_node_id(cls, value: Any) -> Any:
        if isinstance(value, dm.DirectRelationReference):
            return dm.NodeId(value.space, value.external_id)
        elif isinstance(value, tuple) and len(value) == 2 and all(isinstance(item, str) for item in value):
            return dm.NodeId(value[0], value[1])
        elif isinstance(value, list):
            return [cls.as_node_id(item) for item in value]
        return value
    def _to_instances_write(
        self,
        cache: set[tuple[str, str]],
        write_none: bool = False,
        allow_version_increase: bool = False,
    ) -> ResourcesWrite:
        resources = ResourcesWrite()
        if self.as_tuple_id() in cache:
            return resources

        properties: dict[str, Any] = {}

        if self.name is not None:
            properties["name"] = self.name

        if self.model is not None:
            properties["model"] = {
                "space":  self.space if isinstance(self.model, str) else self.model.space,
                "externalId": self.model if isinstance(self.model, str) else self.model.external_id,
            }

        if self.commands is not None:
            properties["commands"] = {
                "space":  self.space if isinstance(self.commands, str) else self.commands.space,
                "externalId": self.commands if isinstance(self.commands, str) else self.commands.external_id,
            }

        if self.source is not None or write_none:
            properties["source"] = self.source

        if self.time_resolution is not None:
            properties["timeResolution"] = {
                "space":  self.space if isinstance(self.time_resolution, str) else self.time_resolution.space,
                "externalId": self.time_resolution if isinstance(self.time_resolution, str) else self.time_resolution.external_id,
            }


        if properties:
            this_node = dm.NodeApply(
                space=self.space,
                external_id=self.external_id,
                existing_version=None if allow_version_increase else self.data_record.existing_version,
                type=as_direct_relation_reference(self.node_type),
                sources=[
                    dm.NodeOrEdgeData(
                        source=self._view_id,
                        properties=properties,
                )],
            )
            resources.nodes.append(this_node)
            cache.add(self.as_tuple_id())



        edge_type = dm.DirectRelationReference("power_ops_types", "ShopOutputTimeSeriesDefinition")
        for output_definition in self.output_definition or []:
            other_resources = DomainRelationWrite.from_edge_to_resources(
                cache,
                start_node=self,
                end_node=output_definition,
                edge_type=edge_type,
                write_none=write_none,
                allow_version_increase=allow_version_increase,
            )
            resources.extend(other_resources)

        edge_type = dm.DirectRelationReference("power_ops_types", "ShopAttributeMapping")
        for attribute_mappings_override in self.attribute_mappings_override or []:
            other_resources = DomainRelationWrite.from_edge_to_resources(
                cache,
                start_node=self,
                end_node=attribute_mappings_override,
                edge_type=edge_type,
                write_none=write_none,
                allow_version_increase=allow_version_increase,
            )
            resources.extend(other_resources)

        if isinstance(self.model, DomainModelWrite):
            other_resources = self.model._to_instances_write(cache)
            resources.extend(other_resources)

        if isinstance(self.commands, DomainModelWrite):
            other_resources = self.commands._to_instances_write(cache)
            resources.extend(other_resources)

        if isinstance(self.time_resolution, DomainModelWrite):
            other_resources = self.time_resolution._to_instances_write(cache)
            resources.extend(other_resources)

        return resources


class ShopScenarioApply(ShopScenarioWrite):
    def __new__(cls, *args, **kwargs) -> ShopScenarioApply:
        warnings.warn(
            "ShopScenarioApply is deprecated and will be removed in v1.0. Use ShopScenarioWrite instead."
            "The motivation for this change is that Write is a more descriptive name for the writing version of the"
            "ShopScenario.",
            UserWarning,
            stacklevel=2,
        )
        return super().__new__(cls)


class ShopScenarioList(DomainModelList[ShopScenario]):
    """List of shop scenarios in the read version."""

    _INSTANCE = ShopScenario

    def as_write(self) -> ShopScenarioWriteList:
        """Convert these read versions of shop scenario to the writing versions."""
        return ShopScenarioWriteList([node.as_write() for node in self.data])

    def as_apply(self) -> ShopScenarioWriteList:
        """Convert these read versions of primitive nullable to the writing versions."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()

    @property
    def model(self) -> ShopModelList:
        from ._shop_model import ShopModel, ShopModelList

        return ShopModelList([item.model for item in self.data if isinstance(item.model, ShopModel)])

    @property
    def commands(self) -> ShopCommandsList:
        from ._shop_commands import ShopCommands, ShopCommandsList

        return ShopCommandsList([item.commands for item in self.data if isinstance(item.commands, ShopCommands)])

    @property
    def time_resolution(self) -> ShopTimeResolutionList:
        from ._shop_time_resolution import ShopTimeResolution, ShopTimeResolutionList

        return ShopTimeResolutionList([item.time_resolution for item in self.data if isinstance(item.time_resolution, ShopTimeResolution)])

    @property
    def output_definition(self) -> ShopOutputTimeSeriesDefinitionList:
        from ._shop_output_time_series_definition import ShopOutputTimeSeriesDefinition, ShopOutputTimeSeriesDefinitionList

        return ShopOutputTimeSeriesDefinitionList([item for items in self.data for item in items.output_definition or [] if isinstance(item, ShopOutputTimeSeriesDefinition)])

    @property
    def attribute_mappings_override(self) -> ShopAttributeMappingList:
        from ._shop_attribute_mapping import ShopAttributeMapping, ShopAttributeMappingList

        return ShopAttributeMappingList([item for items in self.data for item in items.attribute_mappings_override or [] if isinstance(item, ShopAttributeMapping)])


class ShopScenarioWriteList(DomainModelWriteList[ShopScenarioWrite]):
    """List of shop scenarios in the writing version."""

    _INSTANCE = ShopScenarioWrite

    @property
    def model(self) -> ShopModelWriteList:
        from ._shop_model import ShopModelWrite, ShopModelWriteList

        return ShopModelWriteList([item.model for item in self.data if isinstance(item.model, ShopModelWrite)])

    @property
    def commands(self) -> ShopCommandsWriteList:
        from ._shop_commands import ShopCommandsWrite, ShopCommandsWriteList

        return ShopCommandsWriteList([item.commands for item in self.data if isinstance(item.commands, ShopCommandsWrite)])

    @property
    def time_resolution(self) -> ShopTimeResolutionWriteList:
        from ._shop_time_resolution import ShopTimeResolutionWrite, ShopTimeResolutionWriteList

        return ShopTimeResolutionWriteList([item.time_resolution for item in self.data if isinstance(item.time_resolution, ShopTimeResolutionWrite)])

    @property
    def output_definition(self) -> ShopOutputTimeSeriesDefinitionWriteList:
        from ._shop_output_time_series_definition import ShopOutputTimeSeriesDefinitionWrite, ShopOutputTimeSeriesDefinitionWriteList

        return ShopOutputTimeSeriesDefinitionWriteList([item for items in self.data for item in items.output_definition or [] if isinstance(item, ShopOutputTimeSeriesDefinitionWrite)])

    @property
    def attribute_mappings_override(self) -> ShopAttributeMappingWriteList:
        from ._shop_attribute_mapping import ShopAttributeMappingWrite, ShopAttributeMappingWriteList

        return ShopAttributeMappingWriteList([item for items in self.data for item in items.attribute_mappings_override or [] if isinstance(item, ShopAttributeMappingWrite)])

class ShopScenarioApplyList(ShopScenarioWriteList): ...



def _create_shop_scenario_filter(
    view_id: dm.ViewId,
    name: str | list[str] | None = None,
    name_prefix: str | None = None,
    model: str | tuple[str, str] | dm.NodeId | dm.DirectRelationReference | Sequence[str | tuple[str, str] | dm.NodeId | dm.DirectRelationReference] | None = None,
    commands: str | tuple[str, str] | dm.NodeId | dm.DirectRelationReference | Sequence[str | tuple[str, str] | dm.NodeId | dm.DirectRelationReference] | None = None,
    source: str | list[str] | None = None,
    source_prefix: str | None = None,
    time_resolution: str | tuple[str, str] | dm.NodeId | dm.DirectRelationReference | Sequence[str | tuple[str, str] | dm.NodeId | dm.DirectRelationReference] | None = None,
    external_id_prefix: str | None = None,
    space: str | list[str] | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters: list[dm.Filter] = []
    if isinstance(name, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("name"), value=name))
    if name and isinstance(name, list):
        filters.append(dm.filters.In(view_id.as_property_ref("name"), values=name))
    if name_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("name"), value=name_prefix))
    if isinstance(model, str | dm.NodeId | dm.DirectRelationReference) or is_tuple_id(model):
        filters.append(dm.filters.Equals(view_id.as_property_ref("model"), value=as_instance_dict_id(model)))
    if model and isinstance(model, Sequence) and not isinstance(model, str) and not is_tuple_id(model):
        filters.append(dm.filters.In(view_id.as_property_ref("model"), values=[as_instance_dict_id(item) for item in model]))
    if isinstance(commands, str | dm.NodeId | dm.DirectRelationReference) or is_tuple_id(commands):
        filters.append(dm.filters.Equals(view_id.as_property_ref("commands"), value=as_instance_dict_id(commands)))
    if commands and isinstance(commands, Sequence) and not isinstance(commands, str) and not is_tuple_id(commands):
        filters.append(dm.filters.In(view_id.as_property_ref("commands"), values=[as_instance_dict_id(item) for item in commands]))
    if isinstance(source, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("source"), value=source))
    if source and isinstance(source, list):
        filters.append(dm.filters.In(view_id.as_property_ref("source"), values=source))
    if source_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("source"), value=source_prefix))
    if isinstance(time_resolution, str | dm.NodeId | dm.DirectRelationReference) or is_tuple_id(time_resolution):
        filters.append(dm.filters.Equals(view_id.as_property_ref("timeResolution"), value=as_instance_dict_id(time_resolution)))
    if time_resolution and isinstance(time_resolution, Sequence) and not isinstance(time_resolution, str) and not is_tuple_id(time_resolution):
        filters.append(dm.filters.In(view_id.as_property_ref("timeResolution"), values=[as_instance_dict_id(item) for item in time_resolution]))
    if external_id_prefix is not None:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if isinstance(space, str):
        filters.append(dm.filters.Equals(["node", "space"], value=space))
    if space and isinstance(space, list):
        filters.append(dm.filters.In(["node", "space"], values=space))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None


class _ShopScenarioQuery(NodeQueryCore[T_DomainModelList, ShopScenarioList]):
    _view_id = ShopScenario._view_id
    _result_cls = ShopScenario
    _result_list_cls_end = ShopScenarioList

    def __init__(
        self,
        created_types: set[type],
        creation_path: list[QueryCore],
        client: CogniteClient,
        result_list_cls: type[T_DomainModelList],
        expression: dm.query.ResultSetExpression | None = None,
        connection_name: str | None = None,
        connection_type: Literal["reverse-list"] | None = None,
        reverse_expression: dm.query.ResultSetExpression | None = None,
    ):
        from ._shop_attribute_mapping import _ShopAttributeMappingQuery
        from ._shop_commands import _ShopCommandsQuery
        from ._shop_model import _ShopModelQuery
        from ._shop_output_time_series_definition import _ShopOutputTimeSeriesDefinitionQuery
        from ._shop_time_resolution import _ShopTimeResolutionQuery

        super().__init__(
            created_types,
            creation_path,
            client,
            result_list_cls,
            expression,
            dm.filters.HasData(views=[self._view_id]),
            connection_name,
            connection_type,
            reverse_expression,
        )

        if _ShopModelQuery not in created_types:
            self.model = _ShopModelQuery(
                created_types.copy(),
                self._creation_path,
                client,
                result_list_cls,
                dm.query.NodeResultSetExpression(
                    through=self._view_id.as_property_ref("model"),
                    direction="outwards",
                ),
                connection_name="model",
            )

        if _ShopCommandsQuery not in created_types:
            self.commands = _ShopCommandsQuery(
                created_types.copy(),
                self._creation_path,
                client,
                result_list_cls,
                dm.query.NodeResultSetExpression(
                    through=self._view_id.as_property_ref("commands"),
                    direction="outwards",
                ),
                connection_name="commands",
            )

        if _ShopTimeResolutionQuery not in created_types:
            self.time_resolution = _ShopTimeResolutionQuery(
                created_types.copy(),
                self._creation_path,
                client,
                result_list_cls,
                dm.query.NodeResultSetExpression(
                    through=self._view_id.as_property_ref("timeResolution"),
                    direction="outwards",
                ),
                connection_name="time_resolution",
            )

        if _ShopOutputTimeSeriesDefinitionQuery not in created_types:
            self.output_definition = _ShopOutputTimeSeriesDefinitionQuery(
                created_types.copy(),
                self._creation_path,
                client,
                result_list_cls,
                dm.query.EdgeResultSetExpression(
                    direction="outwards",
                    chain_to="destination",
                ),
                connection_name="output_definition",
            )

        if _ShopAttributeMappingQuery not in created_types:
            self.attribute_mappings_override = _ShopAttributeMappingQuery(
                created_types.copy(),
                self._creation_path,
                client,
                result_list_cls,
                dm.query.EdgeResultSetExpression(
                    direction="outwards",
                    chain_to="destination",
                ),
                connection_name="attribute_mappings_override",
            )

        self.space = StringFilter(self, ["node", "space"])
        self.external_id = StringFilter(self, ["node", "externalId"])
        self.name = StringFilter(self, self._view_id.as_property_ref("name"))
        self.source = StringFilter(self, self._view_id.as_property_ref("source"))
        self._filter_classes.extend([
            self.space,
            self.external_id,
            self.name,
            self.source,
        ])

    def list_shop_scenario(self, limit: int = DEFAULT_QUERY_LIMIT) -> ShopScenarioList:
        return self._list(limit=limit)


class ShopScenarioQuery(_ShopScenarioQuery[ShopScenarioList]):
    def __init__(self, client: CogniteClient):
        super().__init__(set(), [], client, ShopScenarioList)



from collections.abc import Callable

import pandas as pd
import datetime
from cognite.client import CogniteClient
from cognite.client.data_classes.data_modeling.ids import NodeId
from cognite.client.data_classes.datapoints import Aggregate
from cognite.client.utils._time import ZoneInfo

from cognite.powerops.client._generated.v1.data_classes._core.constants import DEFAULT_QUERY_LIMIT


class DataPointsAPI:
    def __init__(self, client: CogniteClient, get_node_ids: Callable[[int], list[NodeId]]) -> None:
        self._client = client
        self._get_node_ids = get_node_ids

    def retrieve_dataframe(
        self,
        start: int | str | datetime.datetime | None = None,
        end: int | str | datetime.datetime | None = None,
        aggregates: Aggregate | str | list[Aggregate | str] | None = None,
        granularity: str | None = None,
        timezone: str | datetime.timezone | ZoneInfo | None = None,
        target_unit: str | None = None,
        target_unit_system: str | None = None,
        limit: int | None = None,
        timeseries_limit: int = DEFAULT_QUERY_LIMIT,
        include_outside_points: bool = False,
        ignore_unknown_ids: bool = False,
        include_status: bool = False,
        ignore_bad_datapoints: bool = True,
        treat_uncertain_as_bad: bool = True,
        uniform_index: bool = False,
        include_aggregate_name: bool = True,
        include_granularity_name: bool = False,
    ) -> pd.DataFrame:
        """Get datapoints directly in a pandas dataframe.

        Time series support status codes like Good, Uncertain and Bad. You can read more in the Cognite Data Fusion developer documentation on
        `status codes. <https://developer.cognite.com/dev/concepts/reference/quality_codes/>`_

        Note:
            For many more usage examples, check out the :py:meth:`~DatapointsAPI.retrieve` method which accepts exactly the same arguments.

        Args:
            start (int | str | datetime.datetime | None): Inclusive start. Default: 1970-01-01 UTC.
            end (int | str | datetime.datetime | None): Exclusive end. Default: "now"
            aggregates (Aggregate | str | list[Aggregate | str] | None): Single aggregate or list of aggregates to retrieve. Available options: ``average``, ``continuous_variance``, ``count``, ``count_bad``, ``count_good``, ``count_uncertain``, ``discrete_variance``, ``duration_bad``, ``duration_good``, ``duration_uncertain``, ``interpolation``, ``max``, ``min``, ``step_interpolation``, ``sum`` and ``total_variation``. Default: None (raw datapoints returned)
            granularity (str | None): The granularity to fetch aggregates at. Can be given as an abbreviation or spelled out for clarity: ``s/second(s)``, ``m/minute(s)``, ``h/hour(s)``, ``d/day(s)``, ``w/week(s)``, ``mo/month(s)``, ``q/quarter(s)``, or ``y/year(s)``. Examples: ``30s``, ``5m``, ``1day``, ``2weeks``. Default: None.
            timezone (str | datetime.timezone | ZoneInfo | None): For raw datapoints, which timezone to use when displaying (will not affect what is retrieved). For aggregates, which timezone to align to for granularity 'hour' and longer. Align to the start of the hour, -day or -month. For timezones of type Region/Location, like 'Europe/Oslo', pass a string or ``ZoneInfo`` instance. The aggregate duration will then vary, typically due to daylight saving time. You can also use a fixed offset from UTC by passing a string like '+04:00', 'UTC-7' or 'UTC-02:30' or an instance of ``datetime.timezone``. Note: Historical timezones with second offset are not supported, and timezones with minute offsets (e.g. UTC+05:30 or Asia/Kolkata) may take longer to execute.
            target_unit (str | None): The unit_external_id of the datapoints returned. If the time series does not have a unit_external_id that can be converted to the target_unit, an error will be returned. Cannot be used with target_unit_system.
            target_unit_system (str | None): The unit system of the datapoints returned. Cannot be used with target_unit.
            limit (int | None): Maximum number of datapoints to return for each time series. Default: None (no limit)
            timeseries_limit (int): Maximum number of timeseries to fetch (columns in the dataframe). Default: 5
            include_outside_points (bool): Whether to include outside points. Not allowed when fetching aggregates. Default: False
            ignore_unknown_ids (bool): Whether to ignore missing time series rather than raising an exception. Default: False
            include_status (bool): Also return the status code, an integer, for each datapoint in the response. Only relevant for raw datapoint queries, not aggregates.
            ignore_bad_datapoints (bool): Treat datapoints with a bad status code as if they do not exist. If set to false, raw queries will include bad datapoints in the response, and aggregates will in general omit the time period between a bad datapoint and the next good datapoint. Also, the period between a bad datapoint and the previous good datapoint will be considered constant. Default: True.
            treat_uncertain_as_bad (bool): Treat datapoints with uncertain status codes as bad. If false, treat datapoints with uncertain status codes as good. Used for both raw queries and aggregates. Default: True.
            uniform_index (bool): If only querying aggregates AND a single granularity is used AND no limit is used, specifying `uniform_index=True` will return a dataframe with an equidistant datetime index from the earliest `start` to the latest `end` (missing values will be NaNs). If these requirements are not met, a ValueError is raised. Default: False
            include_aggregate_name (bool): Include 'aggregate' in the column name, e.g. `my-ts|average`. Ignored for raw time series. Default: True
            include_granularity_name (bool): Include 'granularity' in the column name, e.g. `my-ts|12h`. Added after 'aggregate' when present. Ignored for raw time series. Default: False

        Returns:
            pd.DataFrame: A pandas DataFrame containing the requested time series. The ordering of columns is ids first, then external_ids. For time series with multiple aggregates, they will be sorted in alphabetical order ("average" before "max").

        Warning:
            If you have duplicated time series in your query, the dataframe columns will also contain duplicates.

            When retrieving raw datapoints with ``ignore_bad_datapoints=False``, bad datapoints with the value NaN can not be distinguished from those
            missing a value (due to being stored in a numpy array); all will become NaNs in the dataframe.
        """
        node_ids = self._get_node_ids(timeseries_limit)
        if not node_ids:
            return pd.DataFrame()
        return self._client.time_series.data.retrieve_dataframe(
            instance_id=node_ids,
            start=start,
            end=end,
            aggregates=aggregates,
            granularity=granularity,
            timezone=timezone,
            target_unit=target_unit,
            target_unit_system=target_unit_system,
            limit=limit,
            include_outside_points=include_outside_points,
            ignore_unknown_ids=ignore_unknown_ids,
            include_status=include_status,
            ignore_bad_datapoints=ignore_bad_datapoints,
            treat_uncertain_as_bad=treat_uncertain_as_bad,
            uniform_index=uniform_index,
            include_aggregate_name=include_aggregate_name,
            include_granularity_name=include_granularity_name,
            column_names="instance_id",
        )



from __future__ import annotations


from cognite.client import data_modeling as dm

from cognite.powerops.client._generated.v1._api._core import DEFAULT_LIMIT_READ, EdgeAPI, _create_edge_filter
from cognite.powerops.client._generated.v1.data_classes._core import DEFAULT_INSTANCE_SPACE


class ShopModelWithAssetsPowerAssetsAPI(EdgeAPI):
    def list(
            self,
            from_shop_model_with_asset: str | list[str] | dm.NodeId | list[dm.NodeId] | None = None,
            from_shop_model_with_asset_space: str = DEFAULT_INSTANCE_SPACE,
            to_power_asset: str | list[str] | dm.NodeId | list[dm.NodeId] | None = None,
            to_power_asset_space: str = DEFAULT_INSTANCE_SPACE,
            external_id_prefix: str | None = None,
            space: str | list[str] | None = None,
            limit=DEFAULT_LIMIT_READ,
    ) ->dm.EdgeList:
        """List power asset edges of a shop model with asset.

        Args:
            from_shop_model_with_asset: ID of the source shop model with asset.
            from_shop_model_with_asset_space: Location of the shop model with assets.
            to_power_asset: ID of the target power asset.
            to_power_asset_space: Location of the power assets.
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            limit: Maximum number of power asset edges to return. Defaults to 25. Set to -1, float("inf") or None
                to return all items.

        Returns:
            The requested power asset edges.

        Examples:

            List 5 power asset edges connected to "my_shop_model_with_asset":

                >>> from cognite.powerops.client._generated.v1 import PowerOpsModelsV1Client
                >>> client = PowerOpsModelsV1Client()
                >>> shop_model_with_asset = client.shop_model_with_assets.power_assets_edge.list("my_shop_model_with_asset", limit=5)

        """
        filter_ = _create_edge_filter(
            dm.DirectRelationReference("power_ops_core", "ShopModelWithAssets"),

            from_shop_model_with_asset,
            from_shop_model_with_asset_space,
            to_power_asset,
            to_power_asset_space,
            external_id_prefix,
            space,
        )
        return self._list(filter_=filter_, limit=limit)



from __future__ import annotations

import logging
from collections.abc import Callable
from functools import cached_property
from typing import TYPE_CHECKING, Optional, Union

import pandas as pd

from cognite.powerops.client.shop.data_classes.shop_result_files import SHOPLogFile, SHOPResultFile, SHOPYamlFile

if TYPE_CHECKING:
    from cognite.powerops.client.shop.data_classes import SHOPRun

logger = logging.getLogger(__name__)


class SHOPRunResult:
    def __init__(
        self,
        retrieve_objective_function: Callable[[SHOPRun], ObjectiveFunction],
        shop_run: SHOPRun,
        cplex: SHOPLogFile,
        shop_messages: SHOPLogFile,
        post_run: SHOPYamlFile,
    ) -> None:
        self._retrieve_objective_function = retrieve_objective_function
        self._shop_run = shop_run
        self._cplex = cplex
        self._shop_messages = shop_messages
        self._post_run = post_run

    @cached_property
    def files(self) -> dict[str, Optional[SHOPResultFile]]:
        """
        Returns a dictionary of the files associated with the shop run.

        The dictionary keys are:
            - post_run
            - shop_messages
            - cplex

        Returns:
            Dictionary of files associated with the shop run.
        """

        return {"post_run": self._post_run, "shop_messages": self._shop_messages, "cplex": self._cplex}

    @cached_property
    def objective_function(self) -> ObjectiveFunction:
        """
        Returns the objective function of the shop run.

        Returns:
            The objective function of the shop run.
        """
        return self._retrieve_objective_function(self._shop_run)

    def error_message(self) -> str | None:
        """
        Returns the error message of the shop run.

        Returns:
            The error message of the shop run.
        """
        if not self._shop_run.succeeded:
            return "(sample) Error: invalid configuration"
        return None

    @property
    def post_run(self) -> Optional[SHOPYamlFile]:
        """
        Returns the post run yaml file of the shop run.

        Returns:
            The post run yaml file of the shop run.
        """
        return self._post_run

    @property
    def cplex(self) -> SHOPLogFile | None:  # TODO rename to cplex_messages? _logs?
        """
        Returns the cplex log file of the shop run.

        Returns:
            The cplex log file of the shop run.
        """
        return self._cplex

    @property
    def shop_messages(self) -> Optional[SHOPLogFile]:  # TODO rename to _logs?
        """
        Returns the shop messages log file of the shop run.

        Returns:
            The shop messages log file of the shop run.
        """
        return self._shop_messages

    def __repr__(self):
        return f"<SHOPRunResult status={self._shop_run.status}>"


class ObjectiveFunction:
    def __init__(
        self,
        external_id: str,
        data: dict,
        watercourse: str,
        penalty_breakdown: dict,
        field_definitions_df: pd.DataFrame,
    ) -> None:
        self._external_id = external_id
        self._data = data
        self._watercourse = watercourse
        self._penalty_breakdown = penalty_breakdown
        self._field_definitions_df = field_definitions_df

    def __repr__(self) -> str:
        return f"<OBJECTIVE sequence_external_id={self._external_id}>"

    @property
    def watercourse(self) -> str:
        """
        Returns the watercourse of the objective function.

        Returns:
            The watercourse of the objective function.
        """
        return self._watercourse

    @property
    def data(self) -> dict:
        """
        Returns the data of the objective function.

        Returns:
            The data of the objective function.
        """
        return self._data

    @property
    def penalty_breakdown(self) -> dict:
        """
        Returns the penalty breakdown of the objective function.

        Returns:
            The penalty breakdown of the objective function.
        """
        return self._penalty_breakdown

    @property
    def field_definitions_df(self) -> pd.DataFrame:
        """
        The field definitions of the objective function.

        Returns:
            The field definitions of the objective function.
        """
        return self._field_definitions_df

    def data_as_str(self) -> str:
        """
        Returns the data of the objective function as a string representation.

        Returns:
            String representation of the data of the objective function.

        """
        return "\n".join([f"{k}= {v}" for k, v in self.data.items()])

    def penalty_breakdown_as_str(self) -> str:
        """
        Returns the penalty breakdown of the objective function as a string.

        Returns:
            Markdown string of the penalty breakdown.
        """

        # return the penalty breakdown as markdown string some time in future?
        separator = "--------------------------\n"

        def breakdown_inner_dict(key):
            if inner_dict := self.penalty_breakdown.get(key):
                title = f"Breakdown of {key} penalties\n{separator}"
                return title + _inner_string_builder(inner_dict)
            else:
                return ""

        def _inner_string_builder(dictionary: dict[str, Union[str, dict]]) -> str:
            _base = ""
            for key, value in dictionary.items():
                # Assumes only one level of nesting
                if isinstance(value, dict):
                    _base += f" - {key.capitalize()}\n"
                    _base += _inner_string_builder(value)
                else:
                    _base += f"  * {format_dict_key(key)}: {value}\n"
            return _base

        def format_dict_key(key: str) -> str:
            return key.replace("_", " ").replace(".", " ").replace("-", " ").capitalize()

        sum_penalties = self.penalty_breakdown.get("sum_penalties")
        major_penalties = self.penalty_breakdown.get("major_penalties")
        minor_penalties = self.penalty_breakdown.get("minor_penalties")

        penalty_as_string = f"Sum penalties: {sum_penalties} |"
        penalty_as_string += f"Major penalties: {major_penalties} |"
        penalty_as_string += f"Minor penalties: {minor_penalties}\n"
        penalty_as_string += separator.replace("-", "=")

        penalty_as_string += breakdown_inner_dict("minor")
        penalty_as_string += breakdown_inner_dict("major")
        penalty_as_string += breakdown_inner_dict("unknown")

        return penalty_as_string



from cognite.powerops.client.shop.data_classes.shop_case import SHOPCase
from cognite.powerops.client.shop.data_classes.shop_result_files import (
    SHOPLogFile,
    SHOPResultFile,
    SHOPYamlFile,
)
from cognite.powerops.client.shop.data_classes.shop_results import (
    ObjectiveFunction,
    SHOPRunResult,
)
from cognite.powerops.client.shop.data_classes.shop_run import SHOPRun
from cognite.powerops.client.shop.data_classes.shop_run_event import SHOPRunEvent

__all__ = [
    "ObjectiveFunction",
    "SHOPCase",
    "SHOPLogFile",
    "SHOPResultFile",
    "SHOPRun",
    "SHOPRunEvent",
    "SHOPRunResult",
    "SHOPYamlFile",
]

# todo: deprecated, remove custom data classes [next version of powerops-sdk]



from __future__ import annotations

import logging
from pathlib import Path

from cognite.powerops.client import PowerOpsClient
from cognite.powerops.resync.config_to_fdm import ResyncImporter
from cognite.powerops.resync.utils import check_all_linked_sources_exist, get_data_model_write_classes

logger = logging.getLogger(__name__)


def plan(client_configuration: Path, configuration: Path) -> None:
    """Generates data model objects from a resync configuration and prints the plan.

    Initializes a ResyncImporter given the input configuration path. The importer is then used to generate data model
    objects from the configuration. The plan is then printed to the console.

    Args:
        configuration: Path to the resync configuration file.
    """
    client = PowerOpsClient.from_config(client_configuration)

    data_model_classes = get_data_model_write_classes(client.v1)

    resync_importer = ResyncImporter.from_yaml(configuration, data_model_classes, client.cdf)
    resync_data_model_objects, external_ids = resync_importer.to_data_model()

    logger.info(resync_data_model_objects)
    logger.info(f"Generated {len(resync_data_model_objects)} objects")
    logger.info(f"External IDs: {external_ids}")


def apply(client_configuration: Path, configuration: Path, client: PowerOpsClient | None = None) -> None:
    """Generates data model objects from a resync configuration and upserts them to CDF.

    Initializes a ResyncImporter given the input configuration path. The importer is then used to generate data model
    objects from the configuration. The data model objects are then upserted to CDF using the provided client. If no
    client is provided, a client is created from the settings. If the configuration contains file configurations, the
    files are uploaded to CDF before the data model objects are upserted. The function also checks that all linked
    sources exist or are referenced in the data model objects being upserted.

    Args:
        configuration: Path to the resync configuration file.
        client: PowerOpsClient to use for upserting the data model objects. If not provided, a client is created from
            the settings.
    """
    client = client or PowerOpsClient.from_config(client_configuration)

    data_model_classes = get_data_model_write_classes(client.v1)

    resync_importer = ResyncImporter.from_yaml(configuration, data_model_classes, client.cdf)
    resync_data_model_objects, external_ids = resync_importer.to_data_model()

    if resync_importer.file_configuration:
        file_external_ids = resync_importer.file_configuration.upload_files_to_cdf(client.cdf)

    # TODO: check all linked sources exist
    check_all_linked_sources_exist(resync_data_model_objects, file_external_ids, [], external_ids)

    logger.info(resync_data_model_objects)
    client.v1.upsert(resync_data_model_objects, replace=resync_importer.overwrite_data)

    logger.info(f"Upserted {len(resync_data_model_objects)} objects")



from __future__ import annotations

import json
from collections import UserList
from collections.abc import Iterable, Iterator, Sequence
from dataclasses import dataclass, field
from datetime import datetime, timedelta
from typing import Any, Optional, cast, overload

import pandas as pd
from cognite.client import CogniteClient
from cognite.client.data_classes import LabelFilter
from cognite.client.data_classes.events import Event
from cognite.client.utils import datetime_to_ms, ms_to_datetime
from typing_extensions import Self

from cognite.powerops.client.shop.data_classes.shop_case import SHOPCase
from cognite.powerops.client.shop.data_classes.shop_file_reference import SHOPFileReference

try:
    from enum import StrEnum
except ImportError:
    from strenum import StrEnum


class SHOPRunStatus(StrEnum):
    SUCCESS = "success"
    FAILED = "failed"
    IN_PROGRESS = "in_progress"


class SHOPProcessEvents:
    finished: str = "POWEROPS_PROCESS_FINISHED"
    failed: str = "POWEROPS_PROCESS_FAILED"
    started: str = "POWEROPS_PROCESS_STARTED"


class SHOPRunEvent:
    event_type: str = "POWEROPS_SHOP_RUN"
    watercourse: str = "shop:watercourse"
    manual_run: str = "shop:manual_run"
    preprocessor_data: str = "shop:preprocessor_data"
    shop_version: str = "shop_version"
    case_file: str = "cog_shop_case_file"
    shop_files: str = "cog_shop_file_list"
    shopstart: str = "shop:starttime"
    shopend: str = "shop:endtime"
    file_source: str = "shop:file_source"
    plants: str = "shop:plants"
    price_scenario: str = "shop:price_scenario"
    user_id: str = "user:identifier"


@dataclass
class SHOPRun:
    """
    This represents a single SHOP run.

    A SHOP run is represented by an event in CDF. This class is a wrapper around the event.

    Args:
        external_id: The external ID of the SHOP run. This matches the external ID of the event in CDF.
        watercourse: The watercourse of the SHOP run.
        start: The start time of the SHOP run.
        end: The end time of the SHOP run.
        shop_version: The version of SHOP used for the SHOP run.
        user_id: The user ID of the user that triggered the SHOP run.
    """

    external_id: str
    watercourse: str
    data_set_id: int
    start: datetime | None
    end: datetime | None
    shop_version: str
    source: str | None
    _case_file_external_id: str | None
    _shop_files: list[SHOPFileReference]
    _client: CogniteClient = field(repr=False)
    _run_event_types: set[str] = field(init=False, default_factory=set)
    _manual_run: bool = True
    plants: Optional[str] = ""
    price_scenario: Optional[str] = ""

    @classmethod
    def load(cls, event: Event) -> Self:
        """
        Load a SHOP run from an event.

        Args:
            event: The event to load from.

        Returns:

        """
        metadata = event.metadata or {}
        if event.type != SHOPRunEvent.event_type:
            raise ValueError(f"Event {event.external_id} is not a SHOP run event!")

        if event._cognite_client is None:
            raise ValueError(f"Event {event.external_id} is not loaded with a cognite client!")

        if preprocessor_raw := metadata.get(SHOPRunEvent.preprocessor_data, None):
            preprocessor_data = json.loads(preprocessor_raw)
        else:
            preprocessor_data = {}

        return cls(
            external_id=event.external_id,
            data_set_id=event.data_set_id,
            watercourse=metadata.get(SHOPRunEvent.watercourse, ""),
            start=ms_to_datetime(event.start_time) if event.start_time else None,
            end=ms_to_datetime(event.end_time) if event.end_time else None,
            shop_version=preprocessor_data.get(SHOPRunEvent.shop_version, ""),
            _case_file_external_id=preprocessor_data.get(SHOPRunEvent.case_file, {}).get("external_id"),
            _shop_files=[SHOPFileReference.load(item) for item in preprocessor_data.get(SHOPRunEvent.shop_files, [])],
            _client=event._cognite_client,
            source=event.source,
            plants=metadata.get(SHOPRunEvent.plants),
            price_scenario=metadata.get(SHOPRunEvent.price_scenario),
        )

    @property
    def case_file_external_id(self) -> str:
        return self._case_file_external_id

    @property
    def shop_files_external_ids(self) -> list[dict[str, Any]]:
        return [shop_file.dump() for shop_file in self._shop_files]

    def as_cdf_event(self) -> Event:
        shop_version_spec = {SHOPRunEvent.shop_version: self.shop_version} if self.shop_version else {}
        return Event(
            external_id=self.external_id,
            type=SHOPRunEvent.event_type,
            subtype=SHOPRunEvent.event_type,
            data_set_id=self.data_set_id,
            start_time=datetime_to_ms(self.start),
            end_time=datetime_to_ms(self.end) if self.end else None,
            metadata={
                SHOPRunEvent.watercourse: self.watercourse,
                SHOPRunEvent.manual_run: self._manual_run,
                SHOPRunEvent.plants: self.plants,
                SHOPRunEvent.price_scenario: self.price_scenario,
                SHOPRunEvent.preprocessor_data: json.dumps(
                    {
                        **shop_version_spec,
                        SHOPRunEvent.case_file: {"external_id": self._case_file_external_id},
                        SHOPRunEvent.shop_files: [shop_file.dump() for shop_file in self._shop_files],
                    }
                ),
                # These are required by the SHOP container
                # In the functions, create_bid_process_event the end is by default 2 weeks into the future.
                SHOPRunEvent.file_source: self.source,
                SHOPRunEvent.shopstart: self.start.isoformat(),
                SHOPRunEvent.shopend: (self.start + timedelta(days=14)).isoformat(),
            },
            source=self.source,
        )

    def dump(self) -> dict[str, Any]:
        """
        Dump the SHOP run to a dictionary.

        Returns:
            A dictionary representation of the SHOP run.
        """
        return {
            "external_id": self.external_id,
            "watercourse": self.watercourse,
            "start": self.start,
            "end": self.end,
            "case_file_external_id": self.case_file_external_id,
            "shop_files_external_ids": self.shop_files_external_ids,
            "shop_version": self.shop_version,
        }

    def to_case(self):
        """Make a new SHOPCase from this SHOPRun."""
        return SHOPCase(
            data=self.get_case_file(),
            shop_files=self._shop_files.copy(),
            watercourse=self.watercourse.lower(),
            shop_version=self.shop_version,
        )

    def _download_file(self, external_id: str) -> str:
        content_bytes = self._client.files.download_bytes(external_id=external_id)
        try:
            return content_bytes.decode("utf-8")
        except UnicodeDecodeError:
            # SHOP sometimes writes files with Windows-1252 encoding (ASCII)
            # Trying to recover.
            return content_bytes.decode("Windows-1252")

    def get_case_file(self) -> str:
        """
        Get the case file for the SHOP run.

        Returns:
            The case file as a string.
        """
        return self._download_file(self._case_file_external_id)

    def get_shop_files(self) -> Iterable[str]:
        """
        Get the SHOP files for the SHOP run.

        Returns:
            A generator of strings, where each string is a SHOP file.
        """
        for shop_file in self._shop_files:
            yield self._download_file(shop_file.external_id)

    def get_pre_run_file(self) -> str:
        """
        Get the pre-run file generated by CogSHOP. This file is linked to the SHOP run event through the relationship
        "relationship_to.pre_run_file" and contains the input data for the SHOP run.

        Returns:
            The pre-run file as a string.
        """
        pre_run_files = self._client.relationships.list(
            source_external_ids=[self.external_id],
            source_types=["event"],
            target_types=["file"],
            labels=LabelFilter(contains_all=["relationship_to.pre_run_file"]),
        )
        if not pre_run_files:
            raise ValueError(f"No pre-run file found for SHOP run with external ID: {self.external_id}")
        if len(pre_run_files) > 1:
            raise ValueError(f"Multiple pre-run files found for SHOP run with external ID: {self.external_id}")

        return self._download_file(pre_run_files[0].target_external_id)

    def get_log_files(self) -> Iterable[tuple[str, str]]:
        """
        Get the log files for the SHOP run. This is, for exampl,e the CPLEX output.

        Returns:
            A generator of tuples of the form (external_id, file_content).
        """
        if self.check_status() is not SHOPRunStatus.SUCCESS:
            raise ValueError("Cannot retrieve result files for a SHOP run that has not finished successfully.")
        relationship = self._client.relationships.list(source_external_ids=[self.external_id], target_types=["file"])
        for rel in relationship:
            yield rel.target_external_id, self._download_file(rel.target_external_id)

    def check_status(self) -> SHOPRunStatus:
        """
        Check the status of the SHOP run.

        This does a request to CDF to check whether the SHOP run has finished or failed.

        Returns:
            A SHOP run status.

        """
        return self._check_status(update_events=True)

    def _check_status(self, update_events: bool = False) -> SHOPRunStatus:
        # If the run is finished or failed, we don't need to check the events.
        if SHOPProcessEvents.failed in self._run_event_types:
            return SHOPRunStatus.FAILED
        elif SHOPProcessEvents.finished in self._run_event_types:
            return SHOPRunStatus.SUCCESS

        if not update_events:
            return SHOPRunStatus.IN_PROGRESS

        self._update_run_events()
        return self._check_status(update_events=False)

    def get_failure_info(self) -> dict | None:
        relationships = self._client.relationships.list(
            source_external_ids=[self.external_id], target_types=["event"], fetch_resources=True
        )
        failure_events = [rel.target for rel in relationships if rel.target.type == SHOPProcessEvents.failed]
        if not failure_events:
            return None
        failure_events = sorted(failure_events, key=lambda event: -event.created_time)
        cogshop_failure_events = [event for event in failure_events if event.source == "CogShop"]
        argo_failure_events = [event for event in failure_events if event.source == "CogShopExitHandler"]
        error = None
        failures = None
        if cogshop_failure_events:
            error = cogshop_failure_events[0].metadata.get("errorStackTrace")
        if argo_failure_events:
            failures_json = argo_failure_events[0].metadata.get("failures")
            try:
                failures = json.loads(failures_json)
            except (json.JSONDecodeError, TypeError):
                failures = failures_json
        return {
            "error": error,
            "failures": failures,
        }

    def _update_run_events(self) -> None:
        relationships = self._client.relationships.list(
            source_external_ids=[self.external_id], target_types=["event"], fetch_resources=True
        )
        self._run_event_types |= {cast(Event, rel.target).type for rel in relationships}

    def _repr_html_(self) -> str:
        return pd.Series(self.dump()).to_frame().rename(columns={0: "Value"})._repr_html_()

    def __str__(self) -> str:
        """
        Returns a human-readable dump of the SHOP run.
        """
        return json.dumps(self.dump(), indent=4, default=str)


class SHOPRunList(UserList):
    """
    This represents a list of SHOP runs.
    """

    @overload
    def __getitem__(self, item: int) -> SHOPRun: ...

    @overload
    def __getitem__(self, item: slice) -> SHOPRunList: ...

    # The dunder implementations is to get proper type hints
    def __getitem__(self, item: int | slice) -> SHOPRunList | SHOPRun:
        if isinstance(item, slice):
            return type(self)(self.data[item])
        return self.data[item]

    def __iter__(self) -> Iterator[SHOPRun]:
        return super().__iter__()

    @classmethod
    def load(cls, events: Sequence[Event]) -> Self:
        """
        Load a SHOP run list from a list of events.

        Args:
            events: The events to load from. Must be SHOP run events.

        Returns:
            A SHOP run list.
        """
        return cls([SHOPRun.load(event) for event in events])

    def to_pandas(self) -> pd.DataFrame:
        """
        Convert the SHOP run list to a pandas DataFrame.

        Returns:
            A pandas DataFrame.
        """
        return pd.DataFrame([run.dump() for run in self.data])

    def _repr_html_(self) -> str:
        return self.to_pandas()._repr_html_()



from __future__ import annotations

import warnings
from collections.abc import Sequence
from typing import TYPE_CHECKING, Any, ClassVar, Literal,  no_type_check, Optional, Union

from cognite.client import data_modeling as dm, CogniteClient
from cognite.client.data_classes import (
    TimeSeries as CogniteTimeSeries,
    TimeSeriesWrite as CogniteTimeSeriesWrite,
)
from pydantic import Field
from pydantic import field_validator, model_validator

from cognite.powerops.client._generated.v1.data_classes._core import (
    DEFAULT_INSTANCE_SPACE,
    DEFAULT_QUERY_LIMIT,
    DataRecord,
    DataRecordGraphQL,
    DataRecordWrite,
    DomainModel,
    DomainModelWrite,
    DomainModelWriteList,
    DomainModelList,
    DomainRelation,
    DomainRelationWrite,
    GraphQLCore,
    ResourcesWrite,
    FileMetadata,
    FileMetadataWrite,
    FileMetadataGraphQL,
    TimeSeries,
    TimeSeriesWrite,
    TimeSeriesGraphQL,
    T_DomainModelList,
    as_direct_relation_reference,
    as_instance_dict_id,
    as_node_id,
    as_pygen_node_id,
    are_nodes_equal,
    is_tuple_id,
    select_best_node,
    QueryCore,
    NodeQueryCore,
    StringFilter,
    FloatFilter,
    IntFilter,
)
from cognite.powerops.client._generated.v1.data_classes._power_asset import PowerAsset, PowerAssetWrite

if TYPE_CHECKING:
    from cognite.powerops.client._generated.v1.data_classes._generator_efficiency_curve import GeneratorEfficiencyCurve, GeneratorEfficiencyCurveList, GeneratorEfficiencyCurveGraphQL, GeneratorEfficiencyCurveWrite, GeneratorEfficiencyCurveWriteList
    from cognite.powerops.client._generated.v1.data_classes._turbine_efficiency_curve import TurbineEfficiencyCurve, TurbineEfficiencyCurveList, TurbineEfficiencyCurveGraphQL, TurbineEfficiencyCurveWrite, TurbineEfficiencyCurveWriteList


__all__ = [
    "Generator",
    "GeneratorWrite",
    "GeneratorApply",
    "GeneratorList",
    "GeneratorWriteList",
    "GeneratorApplyList",
    "GeneratorFields",
    "GeneratorTextFields",
    "GeneratorGraphQL",
]


GeneratorTextFields = Literal["external_id", "name", "display_name", "asset_type", "start_stop_cost_time_series", "availability_time_series"]
GeneratorFields = Literal["external_id", "name", "display_name", "ordering", "asset_type", "production_min", "penstock_number", "start_stop_cost", "start_stop_cost_time_series", "availability_time_series", "production_max"]

_GENERATOR_PROPERTIES_BY_FIELD = {
    "external_id": "externalId",
    "name": "name",
    "display_name": "displayName",
    "ordering": "ordering",
    "asset_type": "assetType",
    "production_min": "productionMin",
    "penstock_number": "penstockNumber",
    "start_stop_cost": "startStopCost",
    "start_stop_cost_time_series": "startStopCostTimeSeries",
    "availability_time_series": "availabilityTimeSeries",
    "production_max": "productionMax",
}

class GeneratorGraphQL(GraphQLCore):
    """This represents the reading version of generator, used
    when data is retrieved from CDF using GraphQL.

    It is used when retrieving data from CDF using GraphQL.

    Args:
        space: The space where the node is located.
        external_id: The external id of the generator.
        data_record: The data record of the generator node.
        name: Name for the PowerAsset
        display_name: Display name for the PowerAsset.
        ordering: The ordering of the asset
        asset_type: The type of the asset
        production_min: The production min field.
        penstock_number: The penstock number field.
        start_stop_cost: The start stop cost field.
        start_stop_cost_time_series: The start stop cost time series field.
        availability_time_series: The availability time series field.
        generator_efficiency_curve: The generator efficiency curve field.
        production_max: The production max field.
        turbine_efficiency_curves: TODO description
    """
    view_id: ClassVar[dm.ViewId] = dm.ViewId("power_ops_core", "Generator", "1")
    name: Optional[str] = None
    display_name: Optional[str] = Field(None, alias="displayName")
    ordering: Optional[int] = None
    asset_type: Optional[str] = Field(None, alias="assetType")
    production_min: Optional[float] = Field(None, alias="productionMin")
    penstock_number: Optional[int] = Field(None, alias="penstockNumber")
    start_stop_cost: Optional[float] = Field(None, alias="startStopCost")
    start_stop_cost_time_series: Optional[TimeSeriesGraphQL] = Field(None, alias="startStopCostTimeSeries")
    availability_time_series: Optional[TimeSeriesGraphQL] = Field(None, alias="availabilityTimeSeries")
    generator_efficiency_curve: Optional[GeneratorEfficiencyCurveGraphQL] = Field(default=None, repr=False, alias="generatorEfficiencyCurve")
    production_max: Optional[float] = Field(None, alias="productionMax")
    turbine_efficiency_curves: Optional[list[TurbineEfficiencyCurveGraphQL]] = Field(default=None, repr=False, alias="turbineEfficiencyCurves")

    @model_validator(mode="before")
    def parse_data_record(cls, values: Any) -> Any:
        if not isinstance(values, dict):
            return values
        if "lastUpdatedTime" in values or "createdTime" in values:
            values["dataRecord"] = DataRecordGraphQL(
                created_time=values.pop("createdTime", None),
                last_updated_time=values.pop("lastUpdatedTime", None),
            )
        return values
    @field_validator("generator_efficiency_curve", "turbine_efficiency_curves", mode="before")
    def parse_graphql(cls, value: Any) -> Any:
        if not isinstance(value, dict):
            return value
        if "items" in value:
            return value["items"]
        return value

    # We do the ignore argument type as we let pydantic handle the type checking
    @no_type_check
    def as_read(self) -> Generator:
        """Convert this GraphQL format of generator to the reading format."""
        if self.data_record is None:
            raise ValueError("This object cannot be converted to a read format because it lacks a data record.")
        return Generator(
            space=self.space,
            external_id=self.external_id,
            data_record=DataRecord(
                version=0,
                last_updated_time=self.data_record.last_updated_time,
                created_time=self.data_record.created_time,
            ),
            name=self.name,
            display_name=self.display_name,
            ordering=self.ordering,
            asset_type=self.asset_type,
            production_min=self.production_min,
            penstock_number=self.penstock_number,
            start_stop_cost=self.start_stop_cost,
            start_stop_cost_time_series=self.start_stop_cost_time_series.as_read() if self.start_stop_cost_time_series else None,
            availability_time_series=self.availability_time_series.as_read() if self.availability_time_series else None,
            generator_efficiency_curve=self.generator_efficiency_curve.as_read()
if isinstance(self.generator_efficiency_curve, GraphQLCore)
else self.generator_efficiency_curve,
            production_max=self.production_max,
            turbine_efficiency_curves=[turbine_efficiency_curve.as_read() for turbine_efficiency_curve in self.turbine_efficiency_curves or []],
        )


    # We do the ignore argument type as we let pydantic handle the type checking
    @no_type_check
    def as_write(self) -> GeneratorWrite:
        """Convert this GraphQL format of generator to the writing format."""
        return GeneratorWrite(
            space=self.space,
            external_id=self.external_id,
            data_record=DataRecordWrite(existing_version=0),
            name=self.name,
            display_name=self.display_name,
            ordering=self.ordering,
            asset_type=self.asset_type,
            production_min=self.production_min,
            penstock_number=self.penstock_number,
            start_stop_cost=self.start_stop_cost,
            start_stop_cost_time_series=self.start_stop_cost_time_series.as_write() if self.start_stop_cost_time_series else None,
            availability_time_series=self.availability_time_series.as_write() if self.availability_time_series else None,
            generator_efficiency_curve=self.generator_efficiency_curve.as_write()
if isinstance(self.generator_efficiency_curve, GraphQLCore)
else self.generator_efficiency_curve,
            production_max=self.production_max,
            turbine_efficiency_curves=[turbine_efficiency_curve.as_write() for turbine_efficiency_curve in self.turbine_efficiency_curves or []],
        )


class Generator(PowerAsset):
    """This represents the reading version of generator.

    It is used to when data is retrieved from CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the generator.
        data_record: The data record of the generator node.
        name: Name for the PowerAsset
        display_name: Display name for the PowerAsset.
        ordering: The ordering of the asset
        asset_type: The type of the asset
        production_min: The production min field.
        penstock_number: The penstock number field.
        start_stop_cost: The start stop cost field.
        start_stop_cost_time_series: The start stop cost time series field.
        availability_time_series: The availability time series field.
        generator_efficiency_curve: The generator efficiency curve field.
        production_max: The production max field.
        turbine_efficiency_curves: TODO description
    """
    _view_id: ClassVar[dm.ViewId] = dm.ViewId("power_ops_core", "Generator", "1")

    node_type: Union[dm.DirectRelationReference, None] = dm.DirectRelationReference("power_ops_types", "Generator")
    production_min: float = Field(alias="productionMin")
    penstock_number: int = Field(alias="penstockNumber")
    start_stop_cost: float = Field(alias="startStopCost")
    start_stop_cost_time_series: Union[TimeSeries, str, None] = Field(None, alias="startStopCostTimeSeries")
    availability_time_series: Union[TimeSeries, str, None] = Field(None, alias="availabilityTimeSeries")
    generator_efficiency_curve: Union[GeneratorEfficiencyCurve, str, dm.NodeId, None] = Field(default=None, repr=False, alias="generatorEfficiencyCurve")
    production_max: Optional[float] = Field(None, alias="productionMax")
    turbine_efficiency_curves: Optional[list[Union[TurbineEfficiencyCurve, str, dm.NodeId]]] = Field(default=None, repr=False, alias="turbineEfficiencyCurves")

    def as_write(self) -> GeneratorWrite:
        """Convert this read version of generator to the writing version."""
        return GeneratorWrite(
            space=self.space,
            external_id=self.external_id,
            data_record=DataRecordWrite(existing_version=self.data_record.version),
            name=self.name,
            display_name=self.display_name,
            ordering=self.ordering,
            asset_type=self.asset_type,
            production_min=self.production_min,
            penstock_number=self.penstock_number,
            start_stop_cost=self.start_stop_cost,
            start_stop_cost_time_series=self.start_stop_cost_time_series.as_write() if isinstance(self.start_stop_cost_time_series, CogniteTimeSeries) else self.start_stop_cost_time_series,
            availability_time_series=self.availability_time_series.as_write() if isinstance(self.availability_time_series, CogniteTimeSeries) else self.availability_time_series,
            generator_efficiency_curve=self.generator_efficiency_curve.as_write()
if isinstance(self.generator_efficiency_curve, DomainModel)
else self.generator_efficiency_curve,
            production_max=self.production_max,
            turbine_efficiency_curves=[turbine_efficiency_curve.as_write() if isinstance(turbine_efficiency_curve, DomainModel) else turbine_efficiency_curve for turbine_efficiency_curve in self.turbine_efficiency_curves or []],
        )

    def as_apply(self) -> GeneratorWrite:
        """Convert this read version of generator to the writing version."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()

    @classmethod
    def _update_connections(
        cls,
        instances: dict[dm.NodeId | str, Generator],  # type: ignore[override]
        nodes_by_id: dict[dm.NodeId | str, DomainModel],
        edges_by_source_node: dict[dm.NodeId, list[dm.Edge | DomainRelation]],
    ) -> None:
        from ._generator_efficiency_curve import GeneratorEfficiencyCurve
        from ._turbine_efficiency_curve import TurbineEfficiencyCurve

        for instance in instances.values():
            if isinstance(instance.generator_efficiency_curve, (dm.NodeId, str)) and (generator_efficiency_curve := nodes_by_id.get(instance.generator_efficiency_curve)) and isinstance(
                    generator_efficiency_curve, GeneratorEfficiencyCurve
            ):
                instance.generator_efficiency_curve = generator_efficiency_curve
            if edges := edges_by_source_node.get(instance.as_id()):
                turbine_efficiency_curves: list[TurbineEfficiencyCurve | str | dm.NodeId] = []
                for edge in edges:
                    value: DomainModel | DomainRelation | str | dm.NodeId
                    if isinstance(edge, DomainRelation):
                        value = edge
                    else:
                        other_end: dm.DirectRelationReference = (
                            edge.end_node
                            if edge.start_node.space == instance.space
                            and edge.start_node.external_id == instance.external_id
                            else edge.start_node
                        )
                        destination: dm.NodeId | str = (
                            as_node_id(other_end)
                            if other_end.space != DEFAULT_INSTANCE_SPACE
                            else other_end.external_id
                        )
                        if destination in nodes_by_id:
                            value = nodes_by_id[destination]
                        else:
                            value = destination
                    edge_type = edge.edge_type if isinstance(edge, DomainRelation) else edge.type

                    if edge_type == dm.DirectRelationReference("power_ops_types", "isSubAssetOf") and isinstance(
                        value, (TurbineEfficiencyCurve, str, dm.NodeId)
                    ):
                        turbine_efficiency_curves.append(value)

                instance.turbine_efficiency_curves = turbine_efficiency_curves or None




class GeneratorWrite(PowerAssetWrite):
    """This represents the writing version of generator.

    It is used to when data is sent to CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the generator.
        data_record: The data record of the generator node.
        name: Name for the PowerAsset
        display_name: Display name for the PowerAsset.
        ordering: The ordering of the asset
        asset_type: The type of the asset
        production_min: The production min field.
        penstock_number: The penstock number field.
        start_stop_cost: The start stop cost field.
        start_stop_cost_time_series: The start stop cost time series field.
        availability_time_series: The availability time series field.
        generator_efficiency_curve: The generator efficiency curve field.
        production_max: The production max field.
        turbine_efficiency_curves: TODO description
    """
    _view_id: ClassVar[dm.ViewId] = dm.ViewId("power_ops_core", "Generator", "1")

    node_type: Union[dm.DirectRelationReference, dm.NodeId, tuple[str, str], None] = dm.DirectRelationReference("power_ops_types", "Generator")
    production_min: float = Field(alias="productionMin")
    penstock_number: int = Field(alias="penstockNumber")
    start_stop_cost: float = Field(alias="startStopCost")
    start_stop_cost_time_series: Union[TimeSeriesWrite, str, None] = Field(None, alias="startStopCostTimeSeries")
    availability_time_series: Union[TimeSeriesWrite, str, None] = Field(None, alias="availabilityTimeSeries")
    generator_efficiency_curve: Union[GeneratorEfficiencyCurveWrite, str, dm.NodeId, None] = Field(default=None, repr=False, alias="generatorEfficiencyCurve")
    production_max: Optional[float] = Field(None, alias="productionMax")
    turbine_efficiency_curves: Optional[list[Union[TurbineEfficiencyCurveWrite, str, dm.NodeId]]] = Field(default=None, repr=False, alias="turbineEfficiencyCurves")

    @field_validator("generator_efficiency_curve", "turbine_efficiency_curves", mode="before")
    def as_node_id(cls, value: Any) -> Any:
        if isinstance(value, dm.DirectRelationReference):
            return dm.NodeId(value.space, value.external_id)
        elif isinstance(value, tuple) and len(value) == 2 and all(isinstance(item, str) for item in value):
            return dm.NodeId(value[0], value[1])
        elif isinstance(value, list):
            return [cls.as_node_id(item) for item in value]
        return value
    def _to_instances_write(
        self,
        cache: set[tuple[str, str]],
        write_none: bool = False,
        allow_version_increase: bool = False,
    ) -> ResourcesWrite:
        resources = ResourcesWrite()
        if self.as_tuple_id() in cache:
            return resources

        properties: dict[str, Any] = {}

        if self.name is not None:
            properties["name"] = self.name

        if self.display_name is not None or write_none:
            properties["displayName"] = self.display_name

        if self.ordering is not None or write_none:
            properties["ordering"] = self.ordering

        if self.asset_type is not None or write_none:
            properties["assetType"] = self.asset_type

        if self.production_min is not None:
            properties["productionMin"] = self.production_min

        if self.penstock_number is not None:
            properties["penstockNumber"] = self.penstock_number

        if self.start_stop_cost is not None:
            properties["startStopCost"] = self.start_stop_cost

        if self.start_stop_cost_time_series is not None or write_none:
            properties["startStopCostTimeSeries"] = self.start_stop_cost_time_series if isinstance(self.start_stop_cost_time_series, str) or self.start_stop_cost_time_series is None else self.start_stop_cost_time_series.external_id

        if self.availability_time_series is not None or write_none:
            properties["availabilityTimeSeries"] = self.availability_time_series if isinstance(self.availability_time_series, str) or self.availability_time_series is None else self.availability_time_series.external_id

        if self.generator_efficiency_curve is not None:
            properties["generatorEfficiencyCurve"] = {
                "space":  self.space if isinstance(self.generator_efficiency_curve, str) else self.generator_efficiency_curve.space,
                "externalId": self.generator_efficiency_curve if isinstance(self.generator_efficiency_curve, str) else self.generator_efficiency_curve.external_id,
            }

        if self.production_max is not None or write_none:
            properties["productionMax"] = self.production_max


        if properties:
            this_node = dm.NodeApply(
                space=self.space,
                external_id=self.external_id,
                existing_version=None if allow_version_increase else self.data_record.existing_version,
                type=as_direct_relation_reference(self.node_type),
                sources=[
                    dm.NodeOrEdgeData(
                        source=self._view_id,
                        properties=properties,
                )],
            )
            resources.nodes.append(this_node)
            cache.add(self.as_tuple_id())



        edge_type = dm.DirectRelationReference("power_ops_types", "isSubAssetOf")
        for turbine_efficiency_curve in self.turbine_efficiency_curves or []:
            other_resources = DomainRelationWrite.from_edge_to_resources(
                cache,
                start_node=self,
                end_node=turbine_efficiency_curve,
                edge_type=edge_type,
                write_none=write_none,
                allow_version_increase=allow_version_increase,
            )
            resources.extend(other_resources)

        if isinstance(self.generator_efficiency_curve, DomainModelWrite):
            other_resources = self.generator_efficiency_curve._to_instances_write(cache)
            resources.extend(other_resources)

        if isinstance(self.start_stop_cost_time_series, CogniteTimeSeriesWrite):
            resources.time_series.append(self.start_stop_cost_time_series)

        if isinstance(self.availability_time_series, CogniteTimeSeriesWrite):
            resources.time_series.append(self.availability_time_series)

        return resources


class GeneratorApply(GeneratorWrite):
    def __new__(cls, *args, **kwargs) -> GeneratorApply:
        warnings.warn(
            "GeneratorApply is deprecated and will be removed in v1.0. Use GeneratorWrite instead."
            "The motivation for this change is that Write is a more descriptive name for the writing version of the"
            "Generator.",
            UserWarning,
            stacklevel=2,
        )
        return super().__new__(cls)


class GeneratorList(DomainModelList[Generator]):
    """List of generators in the read version."""

    _INSTANCE = Generator

    def as_write(self) -> GeneratorWriteList:
        """Convert these read versions of generator to the writing versions."""
        return GeneratorWriteList([node.as_write() for node in self.data])

    def as_apply(self) -> GeneratorWriteList:
        """Convert these read versions of primitive nullable to the writing versions."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()

    @property
    def generator_efficiency_curve(self) -> GeneratorEfficiencyCurveList:
        from ._generator_efficiency_curve import GeneratorEfficiencyCurve, GeneratorEfficiencyCurveList

        return GeneratorEfficiencyCurveList([item.generator_efficiency_curve for item in self.data if isinstance(item.generator_efficiency_curve, GeneratorEfficiencyCurve)])

    @property
    def turbine_efficiency_curves(self) -> TurbineEfficiencyCurveList:
        from ._turbine_efficiency_curve import TurbineEfficiencyCurve, TurbineEfficiencyCurveList

        return TurbineEfficiencyCurveList([item for items in self.data for item in items.turbine_efficiency_curves or [] if isinstance(item, TurbineEfficiencyCurve)])


class GeneratorWriteList(DomainModelWriteList[GeneratorWrite]):
    """List of generators in the writing version."""

    _INSTANCE = GeneratorWrite

    @property
    def generator_efficiency_curve(self) -> GeneratorEfficiencyCurveWriteList:
        from ._generator_efficiency_curve import GeneratorEfficiencyCurveWrite, GeneratorEfficiencyCurveWriteList

        return GeneratorEfficiencyCurveWriteList([item.generator_efficiency_curve for item in self.data if isinstance(item.generator_efficiency_curve, GeneratorEfficiencyCurveWrite)])

    @property
    def turbine_efficiency_curves(self) -> TurbineEfficiencyCurveWriteList:
        from ._turbine_efficiency_curve import TurbineEfficiencyCurveWrite, TurbineEfficiencyCurveWriteList

        return TurbineEfficiencyCurveWriteList([item for items in self.data for item in items.turbine_efficiency_curves or [] if isinstance(item, TurbineEfficiencyCurveWrite)])

class GeneratorApplyList(GeneratorWriteList): ...



def _create_generator_filter(
    view_id: dm.ViewId,
    name: str | list[str] | None = None,
    name_prefix: str | None = None,
    display_name: str | list[str] | None = None,
    display_name_prefix: str | None = None,
    min_ordering: int | None = None,
    max_ordering: int | None = None,
    asset_type: str | list[str] | None = None,
    asset_type_prefix: str | None = None,
    min_production_min: float | None = None,
    max_production_min: float | None = None,
    min_penstock_number: int | None = None,
    max_penstock_number: int | None = None,
    min_start_stop_cost: float | None = None,
    max_start_stop_cost: float | None = None,
    generator_efficiency_curve: str | tuple[str, str] | dm.NodeId | dm.DirectRelationReference | Sequence[str | tuple[str, str] | dm.NodeId | dm.DirectRelationReference] | None = None,
    min_production_max: float | None = None,
    max_production_max: float | None = None,
    external_id_prefix: str | None = None,
    space: str | list[str] | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters: list[dm.Filter] = []
    if isinstance(name, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("name"), value=name))
    if name and isinstance(name, list):
        filters.append(dm.filters.In(view_id.as_property_ref("name"), values=name))
    if name_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("name"), value=name_prefix))
    if isinstance(display_name, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("displayName"), value=display_name))
    if display_name and isinstance(display_name, list):
        filters.append(dm.filters.In(view_id.as_property_ref("displayName"), values=display_name))
    if display_name_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("displayName"), value=display_name_prefix))
    if min_ordering is not None or max_ordering is not None:
        filters.append(dm.filters.Range(view_id.as_property_ref("ordering"), gte=min_ordering, lte=max_ordering))
    if isinstance(asset_type, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("assetType"), value=asset_type))
    if asset_type and isinstance(asset_type, list):
        filters.append(dm.filters.In(view_id.as_property_ref("assetType"), values=asset_type))
    if asset_type_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("assetType"), value=asset_type_prefix))
    if min_production_min is not None or max_production_min is not None:
        filters.append(dm.filters.Range(view_id.as_property_ref("productionMin"), gte=min_production_min, lte=max_production_min))
    if min_penstock_number is not None or max_penstock_number is not None:
        filters.append(dm.filters.Range(view_id.as_property_ref("penstockNumber"), gte=min_penstock_number, lte=max_penstock_number))
    if min_start_stop_cost is not None or max_start_stop_cost is not None:
        filters.append(dm.filters.Range(view_id.as_property_ref("startStopCost"), gte=min_start_stop_cost, lte=max_start_stop_cost))
    if isinstance(generator_efficiency_curve, str | dm.NodeId | dm.DirectRelationReference) or is_tuple_id(generator_efficiency_curve):
        filters.append(dm.filters.Equals(view_id.as_property_ref("generatorEfficiencyCurve"), value=as_instance_dict_id(generator_efficiency_curve)))
    if generator_efficiency_curve and isinstance(generator_efficiency_curve, Sequence) and not isinstance(generator_efficiency_curve, str) and not is_tuple_id(generator_efficiency_curve):
        filters.append(dm.filters.In(view_id.as_property_ref("generatorEfficiencyCurve"), values=[as_instance_dict_id(item) for item in generator_efficiency_curve]))
    if min_production_max is not None or max_production_max is not None:
        filters.append(dm.filters.Range(view_id.as_property_ref("productionMax"), gte=min_production_max, lte=max_production_max))
    if external_id_prefix is not None:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if isinstance(space, str):
        filters.append(dm.filters.Equals(["node", "space"], value=space))
    if space and isinstance(space, list):
        filters.append(dm.filters.In(["node", "space"], values=space))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None


class _GeneratorQuery(NodeQueryCore[T_DomainModelList, GeneratorList]):
    _view_id = Generator._view_id
    _result_cls = Generator
    _result_list_cls_end = GeneratorList

    def __init__(
        self,
        created_types: set[type],
        creation_path: list[QueryCore],
        client: CogniteClient,
        result_list_cls: type[T_DomainModelList],
        expression: dm.query.ResultSetExpression | None = None,
        connection_name: str | None = None,
        connection_type: Literal["reverse-list"] | None = None,
        reverse_expression: dm.query.ResultSetExpression | None = None,
    ):
        from ._generator_efficiency_curve import _GeneratorEfficiencyCurveQuery
        from ._turbine_efficiency_curve import _TurbineEfficiencyCurveQuery

        super().__init__(
            created_types,
            creation_path,
            client,
            result_list_cls,
            expression,
            dm.filters.HasData(views=[self._view_id]),
            connection_name,
            connection_type,
            reverse_expression,
        )

        if _GeneratorEfficiencyCurveQuery not in created_types:
            self.generator_efficiency_curve = _GeneratorEfficiencyCurveQuery(
                created_types.copy(),
                self._creation_path,
                client,
                result_list_cls,
                dm.query.NodeResultSetExpression(
                    through=self._view_id.as_property_ref("generatorEfficiencyCurve"),
                    direction="outwards",
                ),
                connection_name="generator_efficiency_curve",
            )

        if _TurbineEfficiencyCurveQuery not in created_types:
            self.turbine_efficiency_curves = _TurbineEfficiencyCurveQuery(
                created_types.copy(),
                self._creation_path,
                client,
                result_list_cls,
                dm.query.EdgeResultSetExpression(
                    direction="outwards",
                    chain_to="destination",
                ),
                connection_name="turbine_efficiency_curves",
            )

        self.space = StringFilter(self, ["node", "space"])
        self.external_id = StringFilter(self, ["node", "externalId"])
        self.name = StringFilter(self, self._view_id.as_property_ref("name"))
        self.display_name = StringFilter(self, self._view_id.as_property_ref("displayName"))
        self.ordering = IntFilter(self, self._view_id.as_property_ref("ordering"))
        self.asset_type = StringFilter(self, self._view_id.as_property_ref("assetType"))
        self.production_min = FloatFilter(self, self._view_id.as_property_ref("productionMin"))
        self.penstock_number = IntFilter(self, self._view_id.as_property_ref("penstockNumber"))
        self.start_stop_cost = FloatFilter(self, self._view_id.as_property_ref("startStopCost"))
        self.production_max = FloatFilter(self, self._view_id.as_property_ref("productionMax"))
        self._filter_classes.extend([
            self.space,
            self.external_id,
            self.name,
            self.display_name,
            self.ordering,
            self.asset_type,
            self.production_min,
            self.penstock_number,
            self.start_stop_cost,
            self.production_max,
        ])

    def list_generator(self, limit: int = DEFAULT_QUERY_LIMIT) -> GeneratorList:
        return self._list(limit=limit)


class GeneratorQuery(_GeneratorQuery[GeneratorList]):
    def __init__(self, client: CogniteClient):
        super().__init__(set(), [], client, GeneratorList)
